# ----------------------------------------------------------------------------
# Copyright 2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Defines interface that any backend must implement
"""
import abc
from future.utils import with_metaclass


class Backend_ABC_Meta(abc.ABCMeta):
    """
    metaclass for the backend objects
    takes care of registering all the backend subclasses
    """
    def __init__(self, name, bases, dict_):
        if not hasattr(self, 'backends'):
            self.backends = {}
        else:
            name = getattr(self, 'backend_name', name)
            if name not in ['Backend']:
                self.backends[name] = self
        super(Backend_ABC_Meta, self).__init__(name, bases, dict_)


class AbstractBackend(with_metaclass(Backend_ABC_Meta, object)):
    def __del__(self):
        self.cleanup_backend()

    @abc.abstractmethod
    def cleanup_backend(self):
        """Release any resources that have been acquired by this backend."""
        raise NotImplementedError()

    @abc.abstractmethod
    def gen_rng(self, seed=None):
        """
        Setup the random number generator(s) and store the state
        in self.init_rng_state.

        Arguments:
            seed (int or None): RNG seed, if the seed is None,
                                then a seed will be randomly chosen

        Returns:
            np.random.RandomState: numpy RNG
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def rng_get_state(self, state):
        """
        Get the random number generator state to a specific state.

        Returns a tuple since some backends have multiple RNG states
        (e.g. on-host and on-device)

        Returns:
            tuple: array of numpy ndarray which defines the current
                   state of the RNGs
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def rng_reset(self):
        """
        Reset the random state to the state where the Backend is first
        initialized.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def rng_set_state(self, state):
        """
        Set the random number generator state to a specific state.

        Arguments:
            state (np.array): array which is used to define the RNG
                              state
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def empty(self, shape, dtype=None, name=None, persist_values=True,
              parallel=False, distributed=False):
        """
        Instantiate a new instance of this backend's Tensor class, without
        initializing element values.  This is slightly faster than
        :py:func:`~neon.backends.Backend.array`,
        :py:func:`~neon.backends.Backend.ones`,
        :py:func:`~neon.backends.Backend.zeros`, but the values will be
        random.

        Arguments:
            shape (int, list): length of each dimension of the Tensor.
            dtype (data-type, optional): If present, specifies the underlying
                                         type to employ for each element.
            name (str, optional): name indentifying the tensor (used in printing).
            persist_values (bool, optional): If set to True (the default), the
                                             values assigned to this Tensor
                                             will persist across multiple begin
                                             and end calls.  Setting to False
                                             may provide a performance increase
                                             if values do not need to be
                                             maintained across such calls
            parallel (bool, optional): If True and using multi-GPU backend,
                                       replicate copies of this tensor across
                                       devices.  Defaults to False, and has no
                                       effect on CPU, or (single) GPU backends.
            distributed (bool, optional): If True and using multi-GPU backend,
                                          this tensor is fragmented and
                                          partitioned across devices.  Defaults
                                          to False, and has no effect on CPU,
                                          or (single) GPU backends.

        Returns:
            Tensor: array object

        Raises:
            NotImplementedError: Can't be instantiated directly.

        See Also:
            :py:func:`~neon.backends.backend.Backend.array`,
            :py:func:`~neon.backends.backend.Backend.zeros`,
            :py:func:`~neon.backends.backend.Backend.ones`
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def array(self, ary, dtype=None, name=None, persist_values=True,
              parallel=False, distributed=False):
        """
        Instantiate a new instance of this backend's Tensor class, populating
        elements based on ary values.

        Arguments:
            ary (array_like): input array object to construct from.  Can be
                              built-in python scalar or list (of lists), or a
                              numpy.ndarray
            dtype (data-type, optional): If present, specifies the underlying
                                         type to employ for each element.
            name (str, optional): name indentifying the tensor (used in printing).
            persist_values (bool, optional): If set to True (the default), the
                                             values assigned to this Tensor
                                             will persist across multiple begin
                                             and end calls.  Setting to False
                                             may provide a performance increase
                                             if values do not need to be
                                             maintained across such calls
            parallel (bool, optional): If True and using multi-GPU backend,
                                       replicate copies of this tensor across
                                       devices.  Defaults to False, and has no
                                       effect on CPU, or (single) GPU backends.
            distributed (bool, optional): If True and using multi-GPU backend,
                                          this tensor is fragmented and
                                          partitioned across devices.  Defaults
                                          to False, and has no effect on CPU,
                                          or (single) GPU backends.

        Returns:
            Tensor: array object

        Raises:
            NotImplementedError: Can't be instantiated directly.

        See Also:
            :py:func:`~neon.backends.backend.Backend.empty`,
            :py:func:`~neon.backends.backend.Backend.zeros`,
            :py:func:`~neon.backends.backend.Backend.ones`
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def zeros(self, shape, dtype=None, name=None, persist_values=True,
              parallel=False, distributed=False):
        """
        Instantiate a new instance of this backend's Tensor class, populating
        each element with a value of 0.

        Arguments:
            shape (int, list): length of each dimension of the Tensor.
            dtype (data-type, optional): If present, specifies the underlying
                                         type to employ for each element.
            name (str, optional): name indentifying the tensor (used in printing).
            persist_values (bool, optional): If set to True (the default), the
                                             values assigned to this Tensor
                                             will persist across multiple begin
                                             and end calls.  Setting to False
                                             may provide a performance increase
                                             if values do not need to be
                                             maintained across such calls
            parallel (bool, optional): If True and using multi-GPU backend,
                                       replicate copies of this tensor across
                                       devices.  Defaults to False, and has no
                                       effect on CPU, or (single) GPU backends.
            distributed (bool, optional): If True and using multi-GPU backend,
                                          this tensor is fragmented and
                                          partitioned across devices.  Defaults
                                          to False, and has no effect on CPU,
                                          or (single) GPU backends.

        Returns:
            Tensor: array object

        Raises:
            NotImplementedError: Can't be instantiated directly.

        See Also:
            :py:func:`~neon.backends.backend.Backend.empty`,
            :py:func:`~neon.backends.backend.Backend.ones`,
            :py:func:`~neon.backends.backend.Backend.array`
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def ones(self, shape, dtype=None, name=None, persist_values=True,
             parallel=False, distributed=False):
        """
        Instantiate a new instance of this backend's Tensor class, populating
        each element with a value of 1.

        Arguments:
            shape (int, list): length of each dimension of the Tensor.
            dtype (data-type, optional): If present, specifies the underlying
                                         type to employ for each element.
            name (str, optional): name indentifying the tensor (used in printing).
            persist_values (bool, optional): If set to True (the default), the
                                             values assigned to this Tensor
                                             will persist across multiple begin
                                             and end calls.  Setting to False
                                             may provide a performance increase
                                             if values do not need to be
                                             maintained across such calls
            parallel (bool, optional): If True and using multi-GPU backend,
                                       replicate copies of this tensor across
                                       devices.  Defaults to False, and has no
                                       effect on CPU, or (single) GPU backends.
            distributed (bool, optional): If True and using multi-GPU backend,
                                          this tensor is fragmented and
                                          partitioned across devices.  Defaults
                                          to False, and has no effect on CPU,
                                          or (single) GPU backends.

        Returns:
            Tensor: array object

        Raises:
            NotImplementedError: Can't be instantiated directly.

        See Also:
            :py:func:`~neon.backends.backend.Backend.empty`,
            :py:func:`~neon.backends.backend.Backend.zeros`,
            :py:func:`~neon.backends.backend.Backend.array`
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def empty_like(self, other_ary, name=None, persist_values=True):
        """
        Instantiate a new instance of this backend's Tensor class, with the
        shape taken from other_ary.

        Arguments:
            other_ary (tensor object): Tensor to inherit the dimensions of.
            name (str, optional): name indentifying the tensor (used in printing).
            dtype (data-type, optional): If present, specifies the underlying
                                         type to employ for each element.
            persist_values (bool, optional): If set to True (the default), the
                                             values assigned to this Tensor
                                             will persist across multiple begin
                                             and end calls.  Setting to False
                                             may provide a performance increase
                                             if values do not need to be
                                             maintained across such calls.

        Returns:
            Tensor: array object

        Raises:
            NotImplementedError: Can't be instantiated directly.

        See Also:
            :py:func:`~neon.backends.backend.Backend.empty`,
            :py:func:`~neon.backends.backend.Backend.ones`,
            :py:func:`~neon.backends.backend.Backend.array`
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def zeros_like(self, other_ary, name=None, persist_values=True):
        """
        Instantiate a new instance of this backend's Tensor class, with the
        shape taken from other_ary and populating each element with a value of 0.

        Arguments:
            other_ary (tensor object): Tensor to inherit the dimensions of.
            name (str, optional): name indentifying the tensor (used in printing).
            dtype (data-type, optional): If present, specifies the underlying
                                         type to employ for each element.
            persist_values (bool, optional): If set to True (the default), the
                                             values assigned to this Tensor
                                             will persist across multiple begin
                                             and end calls.  Setting to False
                                             may provide a performance increase
                                             if values do not need to be
                                             maintained across such calls.
        Returns:
            Tensor: array object

        Raises:
            NotImplementedError: Can't be instantiated directly.

        See Also:
            :py:func:`~neon.backends.backend.Backend.empty`,
            :py:func:`~neon.backends.backend.Backend.ones`,
            :py:func:`~neon.backends.backend.Backend.array`
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def dot(self, a, b, out=None):
        """
        Dot product of two Tensors.

        Arguments:
            a (Tensor): left-hand side operand.
            b (Tensor): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.
                                    Note that this object should differ from
                                    left and right.

        Returns:
            OpTreeNode: the resulting op-tree from this operation.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def compound_dot(self, A, B, C, alpha=1.0, beta=0.0, relu=False):
        """
        Perform one of the following operations (* is dot product)
        C = alpha * A * B   + beta * C
        C = alpha * A.T * B + beta * C
        C = alpha * A * B.T + beta * C.

        relu: if true, applied before output (and prior to beta addition)

        The operation will be short-circuited to: out <- alpha * left * right
        if beta has value 0 (the default).

        Arguments:
            A (Tensor): left-hand side operand.
            B (Tensor): right-hand side operand.
            C (Tensor): output operand
            alpha (float. optional): scale A*B term
            beta (float, optional): scale C term before sum
            relu (bool, optional): If True apply ReLu non-linearity before
                                   output.  Defaults to False.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def batched_dot(self, A, B, C, alpha=1.0, beta=0.0, relu=False):
        """
        Perform one of the following operations:
        1 For fprop: A(K, C), B(X,C,N), C(X,K,N) --> call batched_dot(A, B, C)
        2 For bprop: A(K, C), B(X,K,N), C(X,C,N) --> call batched_dot(A.T, B, C)
        3 For update: A(X,K,N), B(X,C,N), C(K,C) --> call batched_dot(A, B.T, C)

        Arguments:
            A (Tensor): left-hand input operand
            B (Tensor): right-hand input operand
            C (Tensor): output operand
            alpha (float. optional): scale A*B term
            beta (float, optional): scale C term before sum
            relu (bool, optional): If True apply ReLu non-linearity before
                                   output.  Defaults to False.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def make_binary_mask(self, out, keepthresh=0.5):
        """
        Create a binary mask for dropout layers.

        Arguments:
            out (Tensor): Output tensor
            keepthresh (float, optional): fraction of ones. Defaults to 0.5
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def add(self, a, b, out=None):
        """
        Perform element-wise addition on the operands, storing the resultant
        values in the out Tensor. Each operand and out must have identical
        shape or be broadcastable as such.

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def subtract(self, a, b, out=None):
        """
        Perform element-wise subtraction on the operands, storing the resultant
        values in the out Tensor. Each operand and out must have identical
        shape or be broadcastable as such.

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def multiply(self, a, b, out=None):
        """
        Perform element-wise multiplication on the operands, storing the
        resultant values in the out Tensor. Each operand and out must have
        identical shape or be broadcastable as such.

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def divide(self, a, b, out=None):
        """
        Perform element-wise division on the operands, storing the
        resultant values in the out Tensor. Each operand and out must have
        identical shape or be broadcastable as such.

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def true_divide(self, a, b, out=None):
        """
        Here it is an alias of divide.
        Instead of the Python traditional 'floor division', this returns a
        true division.

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def power(self, a, b, out=None):
        """
        Perform element-wise raise of tsr values to specified power,
        storing the result in Tensor out. Both Tensor's should have identical
        shape.

        Arguments:
            a (Tensor): input to be transformed.
            b (Tensor, numeric): exponentiated value to be applied to
                                     element.  Examples include 2 (square),
                                     0.5 (sqaure root).
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def reciprocal(self, a, out=None):
        """
        Perform element-wise reciprocal of Tensor `a`, storing the result in
        Tensor out. Both Tensor's should have identical shape.

        Arguments:
            a (Tensor): input to be transformed.
            power (Tensor, numeric): exponentiated value to be applied to
                                     element.  Examples include 2 (square),
                                     0.5 (sqaure root).
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def negative(self, a, out=None):
        """
        Perform element-wise negation of Tensor `a`, storing the result in
        Tensor out. Both Tensor's should have identical shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def sgn(self, a, out=None):
        """
        Perform element-wise indication of the sign of Tensor `a`, storing the
        result in Tensor out. Both Tensor's should have identical shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def absolute(self, a, out=None):
        """
        Perform element-wise absolute value of Tensor `a`, storing the result in
        Tensor out. Both Tensor's should have identical shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def fabs(self, a, out=None):
        """
        Perform element-wise absolute value of Tensor `a`, storing the result
        in Tensor out. Both Tensor's should have identical shape. Implemented as
        an alias of absolute.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def sqrt(self, a, out=None):
        """
        Perform element-wise square-root of Tensor `a`, storing the result in
        Tensor out. Both Tensor's should have identical shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def square(self, a, out=None):
        """
        Perform element-wise square of Tensor `a`, storing the result in Tensor
        out. Both Tensor's should have identical shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def exp(self, a, out=None):
        """
        Perform element-wise exponential transformation on Tensor `a`, storing
        the result in Tensor out. Both Tensor's should have identical shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def exp2(self, a, out=None):
        """
        Perform element-wise 2-based exponential transformation on Tensor `a`,
        storing the result in Tensor out. Both Tensor's should have identical
        shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def safelog(self, a, out=None):
        """
        Perform element-wise natural logarithm transformation on Tensor `a`,
        storing the result in Tensor out. Both Tensor's should have identical
        shape.  This log function has built in safety for underflow.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def log(self, a, out=None):
        """
        Perform element-wise natural logarithm transformation on Tensor `a`,
        storing the result in Tensor out. Both Tensor's should have identical
        shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def log2(self, a, out=None):
        """
        Perform element-wise 2-based logarithm transformation on Tensor `a`,
        storing the result in Tensor out. Both Tensor's should have identical
        shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def sig(self, a, out=None):
        """
        Perform element-wise sigmoid transformation on Tensor `a`,
        storing the result in Tensor out. Both Tensor's should have identical
        shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def sig2(self, a, out=None):
        """
        Perform element-wise 2-based sigmoid logarithm transformation on
        Tensor `a`, storing the result in Tensor out. Both Tensor's should
        have identical shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def tanh(self, a, out=None):
        """
        Perform element-wise hyperbolic tangent transformation on Tensor `a`,
        storing the result in Tensor out. Both Tensor's should have identical
        shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def tanh2(self, a, out=None):
        """
        Perform element-wise 2-based hyperbolic tangent transformation on Tensor
        `a`, storing the result in Tensor out. Both Tensor's should have
        identical shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def finite(self, a, out=None):
        """
        Perform element-wise test of finiteness (not infinity or not Not a
        Number) on Tensor `a`, storing the result in Tensor out. Both Tensor's
        should have identical shape.

        Arguments:
            a (Tensor): input to be transformed.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def equal(self, a, b, out=None):
        """
        Performs element-wise equality testing on each element of left and
        right, storing the result in out. Each operand is assumed to be the
        same shape (or broadcastable as such).

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def not_equal(self, a, b, out=None):
        """
        Performs element-wise non-equality testing on each element of left and
        right, storing the result in out. Each operand is assumed to be the
        same shape (or broadcastable as such).

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def less(self, a, b, out=None):
        """
        Performs element-wise less than testing on each element of left and
        right, storing the result in out. Each operand is assumed to be the
        same shape (or broadcastable as such).

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def less_equal(self, a, b, out=None):
        """
        Performs element-wise less than or equal testing on each element of
        left and right, storing the result in out. Each operand is assumed to
        be the same shape (or broadcastable as such).

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def greater(self, a, b, out=None):
        """
        Performs element-wise greater than testing on each element of left and
        right, storing the result in out. Each operand is assumed to be the
        same shape (or broadcastable as such).

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only theshape op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def greater_equal(self, a, b, out=None):
        """
        Performs element-wise greater than or equal testing on each element of
        left and right, storing the result in out. Each operand is assumed to
        be the same shape (or broadcastable as such).

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def maximum(self, a, b, out=None):
        """
        Performs element-wise maximum value assignment based on corresponding
        elements of left and right, storing the result in out. Each operand is
        assumed to be the same shape (or broadcastable as such).

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def minimum(self, a, b, out=None):
        """
        Performs element-wise minimum value assignment based on corresponding
        elements of left and right, storing the result in out. Each operand is
        assumed to be the same shape (or broadcastable as such).

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def clip(self, a, a_min, a_max, out=None):
        """
        Performs element-wise clipping of Tensor `a`, storing the result in out.
        The clipped value will be between [a_min, a_max].

        Arguments:
            a (Tensor, numeric): left-hand side operand.
            b (Tensor, numeric): right-hand side operand.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def sum(self, a, axis=None, out=None, keepdims=True):
        """
        Calculates the summation of the elements along the specified axis.

        Arguments:
            a (Tensor): the Tensor on which to perform the sum
            axis (int, optional): the dimension along which to compute.
                                  If set to None, we will sum over all
                                  dimensions.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.
            keepdims (bool, optional): Keep the axes being computed over in the
                                       output (with size 1), instead of
                                       collapsing.  Defaults to True.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def max(self, a, axis=None, out=None, keepdims=True):
        """
        Calculates the maximal element value along the specified axes.

        Arguments:
            a (Tensor): the Tensor on which to perform the operation
            axis (int, optional): the dimension along which to compute.
                                  If set to None, we will take max over all
                                  dimensions.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.
            keepdims (bool, optional): Keep the axes being computed over in the
                                       output (with size 1), instead of
                                       collapsing.  Defaults to True.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def min(self, a, axis=None, out=None, keepdims=True):
        """
        Calculates the minimal element value along the specified axes.

        Arguments:
            a (Tensor): the Tensor on which to perform the operation
            axis (int, optional): the dimension along which to compute.
                                  If set to None, we will take min over all
                                  dimensions.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.
            keepdims (bool, optional): Keep the axes being computed over in the
                                       output (with size 1), instead of
                                       collapsing.  Defaults to True.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def argmax(self, a, axis=1, out=None, keepdims=True):
        """
        Calculates the indices of the maximal element value along the specified
        axis.  If multiple elements contain the maximum, only the indices of
        the first are returned.

        Arguments:
            a (Tensor): the Tensor on which to perform the operation
            axis (int, optional): the dimension along which to compute.
                                  If set to None, we will take argmax over all
                                  dimensions.  Defaults to 1
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.
            keepdims (bool, optional): Keep the axes being computed over in the
                                       output (with size 1), instead of
                                       collapsing.  Defaults to True.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def argmin(self, a, axis=1, out=None, keepdims=True):
        """
        Calculates the indices of the minimal element value along the specified
        axis.  If multiple elements contain the minimum, only the indices of
        the first are returned.

        Arguments:
            a (Tensor): the Tensor on which to perform the operation
            axis (int, optional): the dimension along which to compute.
                                  If set to None, we will take argmin over all
                                  dimensions.  Defaults to 1
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.
            keepdims (bool, optional): Keep the axes being computed over in the
                                       output (with size 1), instead of
                                       collapsing.  Defaults to True.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def mean(self, a, axis=None, partial=None, out=None, keepdims=True):
        """
        Calculates the arithmetic mean of the elements along the specified
        axes.

        Arguments:
            a (Tensor): the Tensor on which to perform the operation
            axis (int, optional): the dimension along which to compute.
                                  If set to None, we will take mean over all
                                  dimensions.  Defaults to None
            partial (bool, optional): Not currently used.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.
            keepdims (bool, optional): Keep the axes being computed over in the
                                       output (with size 1), instead of
                                       collapsing.  Defaults to True.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def var(self, a, axis=None, partial=None, out=None, keepdims=True):
        """
        Calculates the variance of the elements along the specified
        axes.

        Arguments:
            a (Tensor): the Tensor on which to perform the operation
            axis (int, optional): the dimension along which to compute.
                                  If set to None, we will take var over all
                                  dimensions.  Defaults to None
            partial (bool, optional): Not currently used.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.
            keepdims (bool, optional): Keep the axes being computed over in the
                                       output (with size 1), instead of
                                       collapsing.  Defaults to True.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def std(self, a, axis=None, partial=None, out=None, keepdims=True):
        """
        Calculates the standard deviation of the elements along the specified
        axes.

        Arguments:
            a (Tensor): the Tensor on which to perform the operation
            axis (int, optional): the dimension along which to compute.
                                  If set to None, we will take std over all
                                  dimensions.
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.
            partial (bool, optional): Not currently used.
            keepdims (bool, optional): Keep the axes being computed over in the
                                       output (with size 1), instead of
                                       collapsing.  Defaults to True.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def take(self, a, indices, axis, out=None):
        """
        Extract elements based on the indices along a given axis.

        Arguments:
            a (Tensor): the Tensor on which to perform the operation
            indices (Tensor, numpy ndarray): indicies of elements to select
            axis (int, optional): the dimension along which to compute.
                                  If set to None, we will extract over all
                                  dimensions (flattened first)
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def onehot(self, indices, axis, out=None):
        """
        Generate optree for converting `indices` to a onehot representation.

        Arguments:
            indices (Tensor): Elements must be of numpy integer type for gpu
                              onehot to work.
            axis (int): the axis along the feature length dimension
            out (Tensor, optional): where the result will be stored. If out is
                                    None, only the op-tree will be returned.

        Returns:
            OpTreeNode: the resulting op-tree
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def update_fc_bias(self, err, out):
        """
        Compute the updated bias gradient for a fully connected network layer.

        Arguments:
            err (Tensor): backpropagated error
            out (Tensor): Where to store the updated gradient value.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def add_fc_bias(self, inputs, bias):
        """
        Add the bias for a fully connected network layer.

        Arguments:
            inputs (Tensor): the input to update.
            bias (Tensor): the amount to increment
        """
        self.ng.add(inputs, bias, out=inputs)

    @abc.abstractmethod
    def conv_layer(self, dtype,
                   N, C, K,
                   D=1, H=1, W=1,
                   T=1, R=1, S=1,
                   pad_d=0, pad_h=0, pad_w=0,
                   str_d=1, str_h=1, str_w=1,
                   relu=False, bsum=False):
        """
        Create a new ConvLayer parameter object.
        This is then passed as an argument to all the convolution operations.

        Arguments:
            dtype (data-type, optional): If present, specifies the underlying
                                         type to employ for each element.

            N (int): Number of images in mini-batch
            C (int): Number of input feature maps
            K (int): Number of output feature maps

            D (int, optional): Depth of input image.  Defaults to 1
            H (int, optional): Height of input image.  Defaults to 1
            W (int, optional): Width of input image.  Defaults to 1

            T (int, optional): Depth of filter kernel.  Defaults to 1
            R (int, optional): Height of filter kernel.  Defaults to 1
            S (int, optional): Width of filter kernel.  Defaults to 1

            pad_d (int, optional): amount of zero-padding around the depth edge
                                   Defaults to 0.
            pad_h (int, optional): amount of zero-padding around the height edge
                                   Defaults to 0.
            pad_w (int, optional): amount of zero-padding around the width edge
                                   Defaults to 0.

            str_d (int, optional): factor to step the filters by in the depth
                                   direction.  Defaults to 1
            str_h (int, optional): factor to step the filters by in the depth
                                   direction.  Defaults to 1
            str_w (int, optional): factor to step the filters by in the depth
                                   direction.  Defaults to 1

            relu (bool, optional): apply a relu transform to the output for
                                   fprop or bprop.  Defaults to False

            bsum (bool, optional): calculate the sum along the batchnorm axis
                                   for fprop or bprop.  Outputs an fp32 tensor
                                   of size Kx1.  Defaults to False.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def fprop_conv(self, layer, I, F, O, alpha=1.0, relu=False, repeat=1):
        """
        Forward propagate the inputs of a convolutional network layer to
        produce output.

        Arguments:
            layer: the conv layer as a parameter object
            I (Tensor): inputs
            F (Tensor): the weights (filters)
            O (Tensor): outputs
            alpha (float, optional): linear scaling.  Defaults to 1.0
            relu (bool, optional): apply ReLu before output.  Default not to.
            repeat (int, optional): Repeat this operation the specified number
                                    of times.  Defaults to 1.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def bprop_conv(self, layer, F, E, grad_I, alpha=1.0, repeat=1):
        """
        Backward propagate the error through a convolutional network layer.

        Arguments:
            layer: the conv layer as a parameter object
            F (Tensor): the weights (filters)
            E (Tensor): errors
            grad_I (Tensor): gradient to inputs (output delta)
            alpha (float, optional): linear scaling.  Defaults to 1.0
            repeat (int, optional): Repeat this operation the specified number
                                    of times.  Defaults to 1.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def update_conv(self, layer, I, E, grad_F, alpha=1.0, repeat=1):
        """
        Compute the updated gradient for a convolutional network layer.

        Arguments:
            layer: the conv layer as a parameter object
            I (Tensor): the inputs
            E (Tensor): the errors
            grad_F (Tensor): filter gradients (weights) to update.
            alpha (float, optional): linear scaling.  Defaults to 1.0
            repeat (int, optional): Repeat this operation the specified number
                                    of times.  Defaults to 1.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def deconv_layer(self, dtype,
                     N, C, K,
                     P, Q,
                     R=1, S=1,
                     pad_d=0, pad_h=0, pad_w=0,
                     str_d=1, str_h=1, str_w=1):
        """
        Create a new Deconvolution parameter object.
        This then is passed as an argument to all deconvolution kernels.

        Arguments:
            dtype (data-type, optional): If present, specifies the underlying
                                         type to employ for each element.

            N (int): Number of images in mini-batch
            C (int): Number of input feature maps
            K (int): Number of output feature maps

            P (int): Height of output
            Q (int): Width of output

            R (int, optional): Height of filter kernel.  Defaults to 1
            S (int, optional): Width of filter kernel.  Defaults to 1

            pad_d (int, optional): amount of zero-padding around the depth edge
                                   Defaults to 0.
            pad_h (int, optional): amount of zero-padding around the height edge
                                   Defaults to 0.
            pad_w (int, optional): amount of zero-padding around the width edge
                                   Defaults to 0.

            str_d (int, optional): factor to step the filters by in the depth
                                   direction.  Defaults to 1
            str_h (int, optional): factor to step the filters by in the depth
                                   direction.  Defaults to 1
            str_w (int, optional): factor to step the filters by in the depth
                                   direction.  Defaults to 1

        Leave spatial dimensions at 1 to allow feature map pooling in the fc layers.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def pool_layer(self, dtype,
                   op, N, C,
                   D=1, H=1, W=1,
                   J=1, T=1, R=1, S=1,
                   pad_j=0, pad_d=0, pad_h=0, pad_w=0,
                   str_j=None, str_d=None, str_h=None, str_w=None):
        """
        Create a new PoolLayer parameter object.
        This then is passed as an argument to all pooling kernels.

        Arguments:
            op (str): "max", "avg", "l2" pooling (currently bprop only supports
                      max, but not avg and l2)
            N (int): Number of images in mini-batch

            C (int): Number of input feature maps
            D (int, optional): Depth of input image.  Defaults to 1
            H (int, optional): Height of input image.  Defaults to 1
            W (int, optional): Width of input image.  Defaults to 1

            J (int, optional): Size of feature map pooling window
                               (maxout n_pieces).  Defaults to 1
            T (int, optional): Depth of pooling window.  Defaults to 1
            R (int, optional): Height of pooling window.  Defaults to 1
            S (int, optional): Width of pooling window.  Defaults to 1

            pad_j (int, optional): amount of zero-padding around the fm pooling
                                   window edge.  Defaults to 0.
            pad_d (int, optional): amount of zero-padding around the depth edge
                                   Defaults to 0.
            pad_h (int, optional): amount of zero-padding around the height edge
                                   Defaults to 0.
            pad_w (int, optional): amount of zero-padding around the width edge
                                   Defaults to 0.

            str_j (int, optional): factor to step the filters by in the fm
                                   pooling window direction.  Defaults to 1
            str_d (int, optional): factor to step the filters by in the depth
                                   direction.  Defaults to 1
            str_h (int, optional): factor to step the filters by in the depth
                                   direction.  Defaults to 1
            str_w (int, optional): factor to step the filters by in the depth
                                   direction.  Defaults to 1

        Leave spatial dimensions at 1 to allow feature map pooling in the fc layers.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def fprop_pool(self, layer, I, O):
        """
        Forward propagate pooling layer.

        Arguments:
            layer (PoolLayer): The pool layer object, different backends have
                               different pool layers.
            I (Tensor): Input tensor.
            O (Tensor): output tensor.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def bprop_pool(self, layer, I, E, grad_I):
        """
        Backward propagate pooling layer.

        Arguments:
            layer (PoolLayer): The pool layer object. Different backends have
                               different pool layers.
            I (Tensor): Input tensor.
            E (Tensor): Error tensor.
            grad_I (Tensor): Gradient tensor (delta)
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def compound_bprop_lut(self, nin, inputs, error, error_t, dW, pad_idx, alpha=1.0, beta=0):
        """
        Backward propagate lookup table layer.

        Arguments:
            nin (int): Number of input word_ids.
            inputs (Tensor): Input tensor.
            error (Tensor): Error tensor.
            error_t (Tensor): Transposed error tensor.
            dW (Tensor): Gradient tensor (delta).
            pad_idx (int):
            alpha (float):
            beta (float):
        """
        raise NotImplementedError()



# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import numpy as np

from neon import NervanaObject
from neon.backends.backend import Tensor


class Initializer(NervanaObject):
    """
    Abstract base class from which parameter tensor initializers inherit.

    Subclasses should implement the ``fill`` method which takes as input a Tensor
    and fills the values based on the initialization scheme.
    """
    def fill(self, param):
        """
        Initialize the provided tensor with values.

        Args:
            param (Tensor): Input Tensor.
        """
        raise NotImplementedError()


class Constant(Initializer):
    """
    Initializes parameters as a constant.
    """
    def __init__(self, val=0.0, name="constantInit"):
        """
        Class constructor.

        Args:
            val (float, optional): The value to assign to all tensor elements
        """
        super(Constant, self).__init__(name=name)
        self.val = val

    def fill(self, param):
        """
        Fills the provided tensor.

        Args:
            param (tensor): target tensor to fill
        """
        if isinstance(self.val, Tensor):
            assert self.val.shape == param.shape, "Constant(Array) initializer can"\
                                                  " only fill a matching shape tensor"
        param[:] = self.val


class Array(Constant):
    """
    Initializes parameters with values specified by a provided numpy array.

    Same functionality as Constant except serialization needs to dump
    tensor values into np array

    Args:
        vals (ndarray or tensor, optional): Values to assign to the tensor elements
    """
    def get_description(self):
        """
        Returns description of the object as a dict. Transfers the
        tensors back to a numpy array.
        """
        desc = super(Array, self).get_description()
        if isinstance(desc['config']['val'], Tensor):
            desc['config']['val'] = desc['config']['val'].get()
        return desc


class Uniform(Initializer):
    """
    Initializes parameters with random values drawn from a uniform distribution.
    """
    def __init__(self, low=0.0, high=1.0, name="uniformInit"):
        """
        Class constructor.

        Args:
            low  (float, optional): Lower bound of range.
            high (float, optional): Upper bound of range.
        """
        super(Uniform, self).__init__(name=name)
        self.low, self.high = (low, high)

    def fill(self, param):
        """
        Fill the provided tensor with random values drawn from a uniform
        distribution.

        Args:
            params (tensor): Tensor to fill
        """
        param[:] = self.be.rng.uniform(self.low, self.high, param.shape)


class Gaussian(Initializer):
    """
    Initializes parameters with a gaussian distribution with the provided mean
    and standard deviation. Defaults to (loc=0, scale=1)
    """
    def __init__(self, loc=0.0, scale=1.0, name="gaussianInit"):
        """
        Class constructor.

        Args:
            loc   (float, optional): Mean parameter (mu). Defaults to 0.
            scale (float, optional): Standard deviation parameter (sigma). Defaults to 1.
            name (string, optional): Name to assign an instance of this class.
        """
        super(Gaussian, self).__init__(name=name)
        self.loc, self.scale = (loc, scale)

    def fill(self, param):
        """
        Fill the provided tensor with random values drawn from a gaussian
        distribution.

        Args:
            params (tensor): Tensor to fill
        """
        param[:] = self.be.rng.normal(self.loc, self.scale, param.shape)


class GlorotUniform(Initializer):
    """
    Initializes parameter tensors with values drawn from a uniform distribution
    ranging from :math:`-K` to :math:`K`. We define :math:`K=\sqrt{6 / (n_{in} + n_{out})}`,
    where :math:`n_{in}` and :math:`n_{out}` are the input and output dimensions, respectively,
    of the parameter tensor. This approach normalizes the range of the initialized values
    by the tensor dimensions.

    From: "Understanding the difficulty of training deep feedforward neural networks"
    (http://jmlr.org/proceedings/papers/v9/glorot10a/glorot10a.pdf).
    """
    def __init__(self, name="autouniformInit"):
        """
        Class constructor.

        Args:
            name (string, optional): Name to assign an instance of this class
        """
        super(GlorotUniform, self).__init__(name=name)

    def fill(self, param):
        """
        Fill the provided tensor with random values drawn from the Uniform
        distribution, using normalized bounds.

        Args:
            params (tensor): Tensor to fill
        """
        k = np.sqrt(6.0 / (param.shape[0] + param.shape[1]))
        param[:] = self.be.rng.uniform(-k, k, param.shape)


class Xavier(Initializer):
    """
    Initializes parameter tensors with values drawn from a uniform distribution
    ranging from :math:`-K` to :math:`K` We define :math:`K=\sqrt{3 / (n_{in})}`,
    where :math:`n_{in}` is the number of input nodes.

    Similar to Glorot except the range is normalized by the input size only.
    """

    def __init__(self, local=True, name="xavier"):
        """
        Class constructor.

        Args:
            local (bool, optional): Whether the layer type is local (Convolutional) or not.
                                      Default is True.
            name (string, optional): Name to assign an instance of this class.
        """
        super(Xavier, self).__init__(name=name)
        self.local = local

    def fill(self, param):
        """
        Fill the provided tensor with random values drawn from the Uniform
        distribution, using normalized bounds.

        Args:
            params (tensor): Tensor to fill
        """
        fan_in = param.shape[0 if self.local else 1]
        scale = np.sqrt(3. / fan_in)
        param[:] = self.be.rng.uniform(-scale, scale, param.shape)


class Kaiming(Initializer):
    """
    Initializes parameters with a zero-mean Gaussian distribution. The standard deviation
    is automatically set as :math:`\sigma=\sqrt{2 / n_{in}}`, where :math:`n_{in}` is
    the input dimension of the tensor.


    Based on the initializer described in: http://arxiv.org/pdf/1502.01852.pdf.
    """
    def __init__(self, local=True, name="Kaiming"):
        """
        Class constructor.

        Args:
            local (bool, optional): Whether the layer type is local (Convolutional) or not.
                                      Default is True.
            name (string, optional): Name to assign an instance of this class.
        """
        super(Kaiming, self).__init__(name=name)
        self.local = local

    def fill(self, param):
        """
        Fill the provided tensor with random values drawn from a gaussian
        distribution.

        Args:
            params (tensor): Tensor to fill
        """
        fan_in = param.shape[0 if self.local else 1]
        scale = np.sqrt(2. / fan_in)
        param[:] = self.be.rng.normal(0, scale, param.shape)


class IdentityInit(Initializer):
    """
    Initializes parameters with the identity matrix.
    """
    def __init__(self, local=True, name="Identity"):
        """
        Class constructor.

        Args:
            local (bool, optional): Whether the layer type is local (Convolutional) or not.
                                      Default is True.
            name (string, optional): Name to assign an instance of this class.
        """
        super(IdentityInit, self).__init__(name=name)
        self.local = local

    def fill(self, param):
        """
        Fill the provided tensor with the identity matrix.

        Args:
            params (tensor): Tensor to fill
        """
        (nin, nout) = param.shape
        w_ary = np.zeros((nin, nout), dtype=np.float32)
        w_ary[:, :nin] = np.eye(nin)
        param[:] = w_ary


class Orthonormal(Initializer):
    """
    Initializes parameters with the single value decomposition of a
    random gaussian matrix.

    Implementation taken from Lasagne. Reference: Saxe et al., http://arxiv.org/abs/1312.6120
    """

    def __init__(self, scale=1.1, name="orthonormal"):
        """
        Class constructor.

        Args:
            scale (float, optional): Scaling factor of values. Defaults to 1.1.
            name (string, optional): Name to assign an instance of this class.
        """
        super(Orthonormal, self).__init__(name=name)
        self.scale = scale

    def fill(self, param):
        """
        Fill the provided tensor using the Orthonormal method.

        Args:
            params (tensor): Tensor to fill
        """
        a = np.random.normal(0.0, 1.0, param.shape)
        u, _, v = np.linalg.svd(a, full_matrices=False)
        # pick the one with the correct shape
        q = u if u.shape == param.shape else v
        param[:] = self.scale * q



# ----------------------------------------------------------------------------
# Copyright 2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from collections import Counter
import numpy as np
import os

from neon import logger as neon_logger
from neon.data.dataiterator import NervanaDataIterator
from neon.data.datasets import Dataset
from neon.util.compat import xrange


class ImageCaption(NervanaDataIterator):
    """
    This class loads in the sentences and CNN image features for image captioning
    that have been taken from Andrej Karpathy's
    `Deep Visual-Semantic Alignments <http://cs.stanford.edu/people/karpathy/deepimagesent/>`_
    work. They are converted to pkl format to avoid using scipy for loading the .mat
    image features.

    The dataset assumes the model takes the precomputed VGG features of an
    image and a sentence converted to its one hot representation. The model
    then transforms both the image and the sentence to the same space and
    prepends the image to the sentence so that it is treated as the first word
    of the sequence to be fed to a RNN.
    """

    end_token = '.'
    image_size = 4096  # Hard code VVG feature size

    def __init__(self, path, max_images=-1):
        """
        Load vocab and image features. Convert sentences to indices

        Args:
            path (str): Directory containing sentences and image features.
            max_images (int): Number of images to load. Set to -1 for max.
        """
        super(ImageCaption, self).__init__(name=None)

        self.path = path
        neon_logger.display('Reading train images and sentences from %s' % self.path)
        self.read_images('train')
        self.load_vocab()

        trainSents, trainImgs = [], []
        for i, img_sent in enumerate(self.iterImageSentencePair()):
            if len(trainSents) > max_images and max_images > 0:
                break
            trainImgs.append(img_sent['image'])
            sent = [self.end_token] + [x for x in img_sent['sentence']['tokens']
                                       if x in self.vocab_to_index]
            trainSents.append(sent[:self.max_sentence_length])

        self.nbatches = len(trainImgs) // self.be.bsz
        self.ndata = self.nbatches * self.be.bsz

        self.X = np.zeros((len(trainSents), self.max_sentence_length))
        self.y = np.zeros((len(trainSents), self.max_sentence_length + 1))
        self.images = np.vstack(trainImgs)

        self.sent_length = np.array([len(x) + 1 for x in trainSents])
        self.sent_ends = np.arange(self.max_sentence_length + 1)[:, np.newaxis]
        for sent_idx, sent in enumerate(trainSents):
            self.X[sent_idx, :len(sent)] = [self.vocab_to_index[word] for word in sent]
        self.y[:, :-1] = self.X

    def load_vocab(self):
        """
        Load vocab and initialize buffers
        Input sentence batch is of dimension (vocab_size, max_sentence_length * batch_size)
        where each column is the 1-hot representation of a word and the first batch_size columns
        are the first words of each sentence.
        """

        sentences = [sent['tokens'] for sent in self.iterSentences()]
        # Flatten list of list of words to one list of words
        words = [word for sentence in sentences for word in sentence]
        # Count words and keep words greater than threshold
        word_counts = Counter(words)

        vocab = [self.end_token] + \
                [word for word in list(word_counts.keys()) if word_counts[word] >= 5]
        self.vocab_size = len(vocab)
        self.vocab_to_index = dict((c, i) for i, c in enumerate(vocab))
        self.index_to_vocab = dict((i, c) for i, c in enumerate(vocab))

        # Compute optional bias vector for initializing final linear layer bias
        word_counts[self.end_token] = len(sentences)
        self.bias_init = np.array([1.0 * word_counts[self.index_to_vocab[i]]
                                   for i in self.index_to_vocab]).reshape((self.vocab_size, 1))
        self.bias_init /= np.sum(self.bias_init)
        self.bias_init = np.log(self.bias_init)
        self.bias_init -= np.max(self.bias_init)

        self.max_sentence_length = max(len(sent) for sent in sentences) + 1

        self.dev_image = self.be.iobuf(self.image_size)
        self.dev_imageT = self.be.empty(self.dev_image.shape[::-1])
        self.dev_X = self.be.iobuf((self.vocab_size, self.max_sentence_length))
        self.dev_y = self.be.iobuf((self.vocab_size, self.max_sentence_length + 1))
        # Create mask to deal with variable length sentences
        self.dev_y_mask = self.be.iobuf((self.vocab_size, self.max_sentence_length + 1))
        self.y_mask = np.zeros(self.dev_y_mask.shape,
                               dtype=np.uint8).reshape(self.vocab_size,
                                                       self.max_sentence_length + 1, -1)
        self.y_mask_reshape = self.y_mask.reshape(self.dev_y_mask.shape)

        self.dev_lbl = self.be.iobuf(self.max_sentence_length, dtype=np.int32)
        self.dev_lblT = self.be.empty(self.dev_lbl.shape[::-1])
        self.dev_lblflat = self.dev_lbl.reshape((1, self.dev_lbl.size))

        self.dev_y_lbl = self.be.iobuf(self.max_sentence_length + 1, dtype=np.int32)
        self.dev_y_lblT = self.be.empty(self.dev_y_lbl.shape[::-1])
        self.dev_y_lblflat = self.dev_y_lbl.reshape((1, self.dev_y_lbl.size))

        self.shape = [self.image_size, (self.vocab_size, self.max_sentence_length)]
        neon_logger.display("Vocab size: %d, Max sentence length: %d" % (self.vocab_size,
                                                                         self.max_sentence_length))

    def read_images(self, split):
        """
        Read sentences and image features from pickled dict

        Args:
            split (str): test or train split
        """
        data_path = os.path.join(self.path, 'features.pkl.gz')
        from neon.util.persist import load_obj
        self.dataset = load_obj(data_path)
        self.sent_data = self.dataset['sents'][split]
        self.features = self.dataset['feats']

    def __iter__(self):
        """
        Generator that can be used to iterate over this dataset.

        Yields:
            tuples, tuples, first tuple contains image features and one hot input sentence
                            second tuple contains one hot target sentence and mask
                            corresponding to 1's up to where each sentence ends and
                            zeros elsewhere after.
        """

        shuf_idx = self.be.rng.permutation(len(self.X))
        self.X, self.y, self.images = (self.X[shuf_idx], self.y[shuf_idx], self.images[shuf_idx])
        self.sent_length = self.sent_length[shuf_idx]

        for batch_idx in xrange(self.nbatches):

            start = batch_idx * self.be.bsz
            end = (batch_idx + 1) * self.be.bsz

            # image_batch = self.images[start:end].T.astype(np.float32, order='C')
            self.dev_imageT.set(self.images[start:end])
            self.dev_image[:] = self.dev_imageT.T

            # X_batch = self.X[start:end].T.astype(np.float32, order='C')
            self.dev_lblT.set(self.X[start:end])
            self.dev_lbl[:] = self.dev_lblT.T
            self.dev_X[:] = self.be.onehot(self.dev_lblflat, axis=0)

            self.y_mask[:] = 1
            sent_lens = self.sent_length[start:end]
            self.y_mask[:, self.sent_ends > sent_lens[np.newaxis, :]] = 0
            self.dev_y_mask[:] = self.y_mask_reshape

            # y_batch = self.y[start:end].T.astype(np.float32, order='C')
            self.dev_y_lblT.set(self.y[start:end])
            self.dev_y_lbl[:] = self.dev_y_lblT.T
            self.dev_y[:] = self.be.onehot(self.dev_y_lblflat, axis=0)
            self.dev_y[:] = self.dev_y * self.dev_y_mask

            yield (self.dev_image, self.dev_X), (self.dev_y, self.dev_y_mask)

    def prob_to_word(self, prob):
        """
        Convert 1 hot probabilities to sentences.

        Args:
            prob (Tensor): Word probabilities of each sentence of batch.
                           Of size (vocab_size, batch_size * (max_sentence_length+1))

        Returns:
            list containing sentences
        """

        sents = []

        if not isinstance(prob, np.ndarray):
            prob = prob.get()
        words = [self.index_to_vocab[x] for x in np.argmax(prob, axis=0).tolist()]

        for sent_index in xrange(self.be.bsz):
            sent = []
            for i in xrange(self.max_sentence_length):
                word = words[self.be.bsz * i + sent_index]
                sent.append(word)
                if (i > 0 and word == self.end_token) or i >= 20:
                    break
            sents.append(" ".join(sent))

        return sents

    def predict(self, model):
        """
        Given a model, generate sentences from this dataset.

        Args:
            model (Model): Image captioning model.

        Returns:
            list, list containing predicted sentences and target sentences
        """
        sents = []
        targets = []
        y = self.be.zeros(self.dev_X.shape)
        for mb_idx, (x, t) in enumerate(self):
            y.fill(0)
            # Repeatedly generate next word in sentence and choose max prob word each time.
            for step in range(1, self.max_sentence_length + 1):
                prob = model.fprop((x[0], y), inference=True).get()[:, :-self.be.bsz].copy()
                pred = np.argmax(prob, axis=0)
                prob.fill(0)
                for i in range(step * self.be.bsz):
                    prob[pred[i], i] = 1
                y[:] = prob
            sents += self.prob_to_word(y)
            # Test set, keep list of targets
            if isinstance(self, ImageCaptionTest):
                targets += t[0]
            # Train set, only 1 target
            else:
                targets.append(t[0])

        return sents, targets

    def bleu_score(self, sents, targets):
        """
        Compute the BLEU score from a list of predicted sentences and reference sentences

        Args:
            sents (list): list of predicted sentences
            targets (list): list of reference sentences where each element is a list of
                            multiple references.
        """

        num_ref = len(targets[0])
        output_file = self.path + '/output'
        reference_files = [self.path + '/reference%d' % i for i in range(num_ref)]
        bleu_script_url = 'https://raw.githubusercontent.com/karpathy/neuraltalk/master/eval/'
        bleu_script = 'multi-bleu.perl'

        neon_logger.display("Writing output and reference sents to dir %s" % self.path)

        output_f = open(output_file, 'w+')
        for sent in sents:
            sent = sent.strip(self.end_token).split()
            output_f.write(" ".join(sent) + '\n')

        reference_f = [open(f, 'w') for f in reference_files]
        for i in range(num_ref):
            for target_sents in targets:
                reference_f[i].write(target_sents[i] + '\n')

        output_f.close()
        [x.close() for x in reference_f]

        owd = os.getcwd()
        os.chdir(self.path)
        if not os.path.exists(bleu_script):
            Dataset.fetch_dataset(bleu_script_url, bleu_script, bleu_script, 6e6)
        bleu_command = 'perl multi-bleu.perl reference < output'
        neon_logger.display("Executing bleu eval script: {}".format(bleu_command))
        os.system(bleu_command)
        os.chdir(owd)

    def _getImage(self, img):
        """
        Get image feature

        Arguments:
            img:

        Returns:

        """
        return self.features[:, img['imgid']]

    def iterSentences(self):
        """Iterate over all sentences"""
        for img in self.sent_data:
            for sent in img['sentences']:
                yield sent

    def iterImageSentencePair(self):
        """Iterate over all image sentence pairs where an image may be repeated"""
        for i, img in enumerate(self.sent_data):
            for sent in img['sentences']:
                out = {}
                out['image'] = self._getImage(img)
                out['sentence'] = sent
                yield out

    def iterImageSentenceGroup(self):
        """Iterate over all image sentence groups"""
        for i, img in enumerate(self.sent_data):
            out = {}
            out['image'] = self._getImage(img)
            out['sentences'] = img['sentences']
            yield out


class ImageCaptionTest(ImageCaption):
    """
    This class loads in image and sentence features for testing.
    """

    def __init__(self, path):
        self.path = path
        neon_logger.display('Reading test images and sentences from %s' % self.path)
        # Load vocab using training set and then load test set
        self.read_images('train')
        self.load_vocab()
        self.read_images('test')

        trainIter = self.iterImageSentenceGroup()
        trainSents, trainImgs = [], []
        for i, img_sent in enumerate(trainIter):
            trainImgs.append(img_sent['image'])
            trainSents.append([' '.join(sent['tokens']) for sent in img_sent['sentences']])

        self.nbatches = len(trainImgs) // self.be.bsz
        self.ndata = self.nbatches * self.be.bsz

        self.images = np.vstack(trainImgs)

        self.ref_sents = trainSents

    def __iter__(self):
        """
        Generator that can be used to iterate over this dataset.

        Yields:
            tuple, tuple: first tuple contains image features and empty input Tensor
                          second tuple contains list of reference sentences and
                          placeholder for mask.
        """
        for batch_idx in xrange(self.nbatches):

            start = batch_idx * self.be.bsz
            end = (batch_idx + 1) * self.be.bsz

            image_batch = self.images[start:end, :].T.astype(np.float32, order='C')
            self.dev_image.set(image_batch)

            yield (self.dev_image, self.dev_X), (self.ref_sents[start:end], None)


class Flickr8k(Dataset):
    """
    Flickr8k data set from http://cs.stanford.edu/people/karpathy/cvpr2015.pdf
    """
    def __init__(self, path='.', max_images=-1):
        url = 'https://s3-us-west-1.amazonaws.com/neon-stockdatasets/image-caption'
        super(Flickr8k, self).__init__('flickr8k.zip',
                                       url,
                                       49165563,
                                       path=path)
        self.max_images = max_images

    def gen_iterators(self):
        data_path = self.load_data()
        self._data_dict = {'train': ImageCaption(path=data_path, max_images=self.max_images)}
        self._data_dict['test'] = ImageCaptionTest(path=data_path)
        return self._data_dict

    def load_data(self):
        return self.load_zip(self.filename, self.size)


class Flickr30k(Dataset):
    """
    Flickr30k data set from http://cs.stanford.edu/people/karpathy/cvpr2015.pdf
    """
    def __init__(self, path='.', max_images=-1):
        url = 'https://s3-us-west-1.amazonaws.com/neon-stockdatasets/image-caption'
        super(Flickr30k, self).__init__('flickr30k.zip',
                                        url,
                                        49165563,
                                        path=path)
        self.max_images = max_images

    def gen_iterators(self):
        data_path = self.load_data()
        self._data_dict = {'train': ImageCaption(path=data_path, max_images=self.max_images)}
        self._data_dict['test'] = ImageCaptionTest(path=data_path)
        return self._data_dict

    def load_data(self):
        return self.load_zip(self.filename, self.size)


class Coco(Dataset):
    """
    MSCOCO data set from http://cs.stanford.edu/people/karpathy/cvpr2015.pdf
    """
    def __init__(self, path='.', max_images=-1):
        url = 'https://s3-us-west-1.amazonaws.com/neon-stockdatasets/image-caption'
        super(Coco, self).__init__('coco.zip',
                                   url,
                                   738051031,
                                   path=path)
        self.max_images = max_images

    def load_data(self):
        return self.load_zip(self.filename, self.size)

    def gen_iterators(self):
        data_path = self.load_data()
        self._data_dict = {'train': ImageCaption(path=data_path, max_images=self.max_images)}
        self._data_dict['test'] = ImageCaptionTest(path=data_path)
        return self._data_dict



# ----------------------------------------------------------------------------
# Copyright 2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from future.utils import native_str
from pycuda.tools import context_dependent_memoize

from neon.backends.cuda_templates import _ew_types
from neon.backends.util.source_module import SourceModule

"""
CUDA kernels for lookup table layers. Kernels are only given for bprop, since
fprop is just a take operation. There is a deterministic kernel and
non-deterministic kernel (using atomics) provided. Sorting kernels are also
provided to help with the deterministic version.
"""


@context_dependent_memoize
def _get_lut_bprop_kernel(dtype, deterministic=False):
    """
    Builds the bprop kernel for lookup table layers based on templated code.
    If the deterministic version is requested, an index buffer must be passed
    as an argument. This index buffer re-orders items in the input tensor
    so that word_ids are sorted. This is required since we need to be sure that
    each thread only updates weights for one word id.

    Arguments:
        dtype (np.dtype): The data which the kernel will operate on.
        deterministic (boolean): Builds the deterministic kernel when this is
            set to True.
    """
    if not deterministic:
        code = r"""
__global__ void lut_bprop(
    int* inputs, %(type)s* dW, %(type)s* errors, const int nin,
    const int embedding_dim, const int vocab_size, const int pad_idx)
{
    const int tid  = threadIdx.x;
    const int bid  = blockIdx.x;

    int word_id = inputs[bid];
    int error_row = bid * embedding_dim;
    int output_row = word_id * embedding_dim;

    if(word_id != pad_idx)
    {
        for(int i = tid; i < embedding_dim; i += blockDim.x)
        {
            atomicAdd(&dW[output_row + i], errors[error_row + i]);
        }
    }
}
"""

        code = code % {
            "type": _ew_types[dtype]["type"]
        }

        module = SourceModule(code, options=["--use_fast_math"])
        kernel = module.get_function("lut_bprop")
        kernel.prepare("PPPIIIi")
    else:
        code = r"""
__global__ void lut_bprop(
    int* inputs, int* index_buffer, %(type)s* dW, %(type)s* errors,
    const int nin, const int embedding_dim, const int vocab_size,
    const int pad_idx)
{
    const int tid  = threadIdx.x;
    const int bid  = blockIdx.x;

    int index_position = bid;
    int index = index_buffer[index_position];
    int word_id = inputs[index];

    if((bid == 0 || word_id != inputs[index_buffer[bid - 1]]) && word_id != pad_idx)
    {
        int output_row = word_id * embedding_dim;

        do {
            int error_row = index * embedding_dim;

            for(int i = tid; i < embedding_dim; i += blockDim.x)
            {
                dW[output_row + i] += errors[error_row + i];
            }

            index_position++;
            if(index_position == gridDim.x)
            {
                break;
            }
            index = index_buffer[index_position];
        } while(inputs[index] == word_id);
    }
}
"""

        code = code % {
            "type": _ew_types[dtype]["type"]
        }

        module = SourceModule(code, options=["--use_fast_math"])
        kernel = module.get_function("lut_bprop")
        kernel.prepare("PPPPIIIi")

    kernel.name = "lut_bprop"
    return kernel


def _get_sorting_kernel(kernel_id, block_size):
    """
    Builds kernels used for sorting inputs. There are several kernels here
    corresponding to the steps in the algorithm. The algorithm works by
    determining the sorted position for each input item. This is done with
    a bucket sort algorithm, where each word_id is a bucket. The first step
    determines the size of each bucket (number of occurences of each word_id).
    Next, a prefix some is computed over the list of bucket sizes to find
    where each bucket will be placed in the output buffer. Finally, each thread
    places it's index into the correct sorted position based on the bucket
    start index (computed from the prefix sum) and that thread's offset into
    the bucket (which is taken from the output of the atomic add done in the
    first step.)

    Arguments:
        kernel_id (Integer): Which step to build the kernel for [0, 4]
        block_size (Integer): Number of threads per block for the prefix sum
            kernels.
    """
    code = r"""
#define THREADS %(threads)s
#define STORE_BLOCKSUM %(store_blocksum)s
__global__ void sort_inputs0(
        int* inputs, int* index_buffer, int* offset_buffer, int* word_counts, const int vocab_size,
        const int input_length)
{
    const int tid = threadIdx.x + (blockDim.x * blockIdx.x);
    int word_id;

    if(tid < input_length)
    {
        word_id = inputs[tid];
        offset_buffer[tid] = atomicAdd(&word_counts[word_id], 1);
    }
}

__device__ void scan(int* buffer, int* blocksum, int global_length)
{
    const int tid = (threadIdx.x << 1) + 1;
    const int gid = ((threadIdx.x + (blockIdx.x * blockDim.x)) << 1) + 1;

    __shared__ int local_counts[THREADS * 2];
    local_counts[tid] = buffer[gid];
    local_counts[tid - 1] = buffer[gid - 1];

    #pragma unroll
    for(int skip = 1; skip <= THREADS; skip <<= 1)
    {
        int mask = (skip << 1) - 1;
        if((tid & mask) == mask)
        {
            local_counts[tid] += local_counts[tid - skip];
        }

        __syncthreads();
    }

    if(tid == (THREADS * 2 - 1))
    {
#if STORE_BLOCKSUM
        blocksum[blockIdx.x] = local_counts[tid];
#endif
        local_counts[tid] = 0;
    }

    #pragma unroll
    for(int skip = THREADS; skip > 0; skip >>= 1)
    {
        int mask = (skip << 1) - 1;
        if((tid & mask) == mask)
        {
            int temp = local_counts[tid - skip];
            local_counts[tid - skip] = local_counts[tid];
            local_counts[tid] += temp;
        }

        __syncthreads();
    }

    if(gid < global_length)
    {
        buffer[gid] = local_counts[tid];
        buffer[gid - 1] = local_counts[tid - 1];
    }
}

__global__ void sort_inputs1(
        int* inputs, int* index_buffer, int* offset_buffer, int* word_counts, const int vocab_size,
        const int input_length)
{
    scan(word_counts, word_counts + vocab_size, vocab_size);
}

__global__ void sort_inputs2(
        int* inputs, int* index_buffer, int* offset_buffer, int* word_counts, const int vocab_size,
        const int input_length)
{
    scan(word_counts + vocab_size, 0, blockDim.x);
}

__global__ void sort_inputs3(
        int* inputs, int* index_buffer, int* offset_buffer, int* word_counts, const int vocab_size,
        const int input_length)
{
    const int gid = (threadIdx.x + (blockIdx.x * blockDim.x)) << 1;

    if(gid < vocab_size)
    {
        word_counts[gid] += word_counts[vocab_size + blockIdx.x];
        word_counts[gid + 1] += word_counts[vocab_size + blockIdx.x];
    }
}

__global__ void sort_inputs4(
        int* inputs, int* index_buffer, int* offset_buffer, int* word_counts, const int vocab_size,
        const int input_length)
{
    const int tid = threadIdx.x + (blockDim.x * blockIdx.x);
    int word_id;

    if(tid < input_length)
    {
        word_id = inputs[tid];
        int sorted_position = word_counts[word_id] + offset_buffer[tid];
        index_buffer[sorted_position] = tid;
    }
}
"""
    code = code % {
        "threads": block_size,
        "store_blocksum": (1 if kernel_id == 1 else 0)
    }
    module = SourceModule(code, options=["--use_fast_math"])

    function_name = "sort_inputs" + native_str(kernel_id)
    kernel = module.get_function(function_name)
    kernel.prepare("PPPPII")
    kernel.name = "sort_inputs"
    return kernel



# ----------------------------------------------------------------------------
# Copyright 2014-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Defines text datatset handling.
"""

import logging
import numpy as np
import os
import h5py

from neon.data.dataiterator import NervanaDataIterator, ArrayIterator
from neon.data.datasets import Dataset
from neon.data.text_preprocessing import pad_sentences, pad_data

logger = logging.getLogger(__name__)


class Text(NervanaDataIterator):
    """
    This class defines methods for loading and iterating over text datasets.
    """

    def __init__(self, time_steps, path, vocab=None, tokenizer=None,
                 onehot_input=True, reverse_target=False, get_prev_target=False):
        """
        Construct a text dataset object.

        Arguments:
            time_steps (int) : Length of a sequence.
            path (str) : Path to text file.
            vocab (python.set) : A set of unique tokens.
            tokenizer (function) : Tokenizer function.
            onehot_input (boolean): One-hot representation of input
            reverse_target (boolean): for sequence to sequence models, set to
                                      True to reverse target sequence. Also
                                      disables shifting target by one.
            get_prev_target (boolean): for sequence to sequence models, set to
                                       True for training data to provide correct
                                       target from previous time step as decoder
                                       input. If condition, shape will be a tuple
                                       of shapes, corresponding to encoder and
                                       decoder inputs.
        """
        super(Text, self).__init__(name=None)

        self.seq_length = time_steps
        self.onehot_input = onehot_input
        self.batch_index = 0
        self.reverse_target = reverse_target
        self.get_prev_target = get_prev_target

        X, y = self._get_data(path, tokenizer, vocab)

        # reshape to preserve sentence continuity across batches
        self.X = X.reshape(self.be.bsz, self.nbatches, time_steps)
        self.y = y.reshape(self.be.bsz, self.nbatches, time_steps)

        # stuff below this comment needs to be cleaned up and commented
        self.nout = self.nclass
        if self.onehot_input:
            self.shape = (self.nout, time_steps)
            self.dev_X = self.be.iobuf((self.nout, time_steps))
            if self.get_prev_target:
                self.dev_Z = self.be.iobuf((self.nout, time_steps))
        else:
            self.shape = (time_steps, 1)
            self.dev_X = self.be.iobuf(time_steps, dtype=np.int32)
            if self.get_prev_target:
                self.dev_Z = self.be.iobuf(time_steps, dtype=np.int32)
        self.decoder_shape = self.shape

        self.dev_y = self.be.iobuf((self.nout, time_steps))
        self.dev_lbl = self.be.iobuf(time_steps, dtype=np.int32)
        self.dev_lblflat = self.dev_lbl.reshape((1, -1))

    def _get_data(self, path, tokenizer, vocab):

        text = open(path).read()
        tokens = self.get_tokens(text, tokenizer)

        # make this a static method
        extra_tokens = len(tokens) % (self.be.bsz * self.seq_length)
        if extra_tokens:
            tokens = tokens[:-extra_tokens]
        self.nbatches = len(tokens) // (self.be.bsz * self.seq_length)
        self.ndata = self.nbatches * self.be.bsz  # no leftovers

        self.vocab = sorted(self.get_vocab(tokens, vocab))
        self.nclass = len(self.vocab)

        # vocab dicts
        self.token_to_index = dict((t, i) for i, t in enumerate(self.vocab))
        self.index_to_token = dict((i, t) for i, t in enumerate(self.vocab))

        # map tokens to indices
        X = np.asarray([self.token_to_index[t] for t in tokens], dtype=np.uint32)
        if self.reverse_target:
            y = X.copy()
        else:
            y = np.concatenate((X[1:], X[:1]))

        return X, y

    @staticmethod
    def create_valid_file(path, valid_split=0.1):
        """
        Create separate files for training and validation.

        Arguments:
            path(str): Path to data file.
            valid_split(float, optional): Fraction of data to set aside for validation.

        Returns:
            str, str : Paths to train file and validation file
        """
        text = open(path).read()

        # create train and valid paths
        filename, ext = os.path.splitext(path)
        train_path = filename + '_train' + ext
        valid_path = filename + '_valid' + ext

        # split data
        train_split = int(len(text) * (1 - valid_split))
        train_text = text[:train_split]
        valid_text = text[train_split:]

        # write train file
        with open(train_path, 'w') as train_file:
            train_file.write(train_text)

        # write valid file
        with open(valid_path, 'w') as valid_file:
            valid_file.write(valid_text)

        return train_path, valid_path

    @staticmethod
    def get_tokens(string, tokenizer=None):
        """
        Map string to a list of tokens.

        Arguments:
            string(str): String to be tokenized.
            token(object): Tokenizer object.
            tokenizer (function) : Tokenizer function.

        Returns:
            list : A list of tokens
        """
        # (if tokenizer is None, we have a list of characters)
        if tokenizer is None:
            return string
        else:
            return tokenizer(string)

    @staticmethod
    def get_vocab(tokens, vocab=None):
        """
        Construct vocabulary from the given tokens.

        Arguments:
            tokens(list): List of tokens.
            vocab:  (Default value = None)

        Returns:
            python.set : A set of unique tokens
        """
        # (if vocab is not None, we check that it contains all tokens)
        if vocab is None:
            return set(tokens)
        else:
            vocab = set(vocab)
            assert vocab >= set(tokens), "the predefined vocab must contain all the tokens"
            return vocab

    @staticmethod
    def pad_sentences(sentences, sentence_length=None, dtype=np.int32, pad_val=0.):
        """
        Deprecated, use neon.data.text_preprocessing.pad_sentences.
        """
        logger.error('pad_sentences in the Text class is deprecated.  This function '
                     'is now in neon.data.text_preprocessing.')
        return pad_sentences(sentences,
                             sentence_length=sentence_length,
                             dtype=dtype,
                             pad_val=pad_val)

    @staticmethod
    def pad_data(path, vocab_size=20000, sentence_length=100, oov=2,
                 start=1, index_from=3, seed=113, test_split=0.2):
        """
        Deprecated, use neon.data.text_preprocessing.pad_data.
        """
        logger.error('pad_data in the Text class is deprecated.  This function'
                     'is now in neon.data.text_preprocessing')
        return pad_data(path,
                        vocab_size=vocab_size,
                        sentence_length=sentence_length,
                        oov=oov,
                        start=start,
                        index_from=index_from,
                        seed=seed,
                        test_split=test_split)

    def reset(self):
        """
        Reset the starting index of this dataset back to zero.
        Relevant for when one wants to call repeated evaluations on the dataset
        but don't want to wrap around for the last uneven minibatch
        Not necessary when ndata is divisible by batch size
        """
        self.batch_index = 0

    def __iter__(self):
        """
        Generator that can be used to iterate over this dataset.

        Yields:
            tuple : the next minibatch of data.
        """
        self.batch_index = 0
        while self.batch_index < self.nbatches:
            X_batch = self.X[:, self.batch_index, :].T.astype(np.float32, order='C')
            if self.reverse_target is False:
                y_batch = self.y[:, self.batch_index, :].T.astype(np.float32, order='C')
            else:
                # reverse target sequence
                y_batch = self.y[:, self.batch_index, ::-1].T.astype(np.float32, order='C')

            self.dev_lbl.set(y_batch)
            self.dev_y[:] = self.be.onehot(self.dev_lblflat, axis=0)

            if self.onehot_input:
                self.dev_lbl.set(X_batch)
                self.dev_X[:] = self.be.onehot(self.dev_lblflat, axis=0)
                if self.get_prev_target:
                    self.dev_Z[:, self.be.bsz:] = self.dev_y[:, :-self.be.bsz]
                    self.dev_Z[:, 0:self.be.bsz] = 0  # zero-hot, no input
            else:
                self.dev_X.set(X_batch)
                if self.get_prev_target:
                    self.dev_lbl.set(y_batch)
                    self.dev_Z[1:, :] = self.dev_lbl[:-1, :]
                    self.dev_Z[0, :] = 0

            self.batch_index += 1

            if self.get_prev_target:
                yield (self.dev_X, self.dev_Z), self.dev_y
            else:
                yield self.dev_X, self.dev_y


class TextNMT(Text):
    """
    Datasets for neural machine translation on French / English bilingual datasets.
    Available at http://www-lium.univ-lemans.fr/~schwenk/cslm_joint_paper/data/bitexts.tgz

    Arguments:
        time_steps (int) : Length of a sequence.
        path (str) : Path to text file.
        tokenizer (function) : Tokenizer function.
        onehot_input (boolean): One-hot representation of input
        get_prev_target (boolean): for sequence to sequence models, set to
                               True for training data to provide correct
                               target from previous time step as decoder
                               input. If condition, shape will be a tuple
                               of shapes, corresponding to encoder and
                               decoder inputs.
        split (str): "train" or "valid" split of the dataset
        dataset (str): 'un2000' for the United Nations dataset or 'eurparl7'
                       for the European Parliament datset.
        subset_pct (float): Percentage of the dataset to use (100 is the full dataset)
    """
    def __init__(self, time_steps, path, tokenizer=None,
                 onehot_input=False, get_prev_target=False, split=None,
                 dataset='un2000', subset_pct=100):
        """
        Load French and English sentence data from file.
        """
        assert dataset in ('europarl7', 'un2000'), "invalid dataset"
        processed_file = os.path.join(path, dataset + '-' + split + '.h5')
        assert os.path.exists(processed_file), "Dataset at '" + processed_file + "' not found"
        self.subset_pct = subset_pct

        super(TextNMT, self).__init__(time_steps, processed_file, vocab=None, tokenizer=tokenizer,
                                      onehot_input=onehot_input, get_prev_target=get_prev_target,
                                      reverse_target=True)

    def _get_data(self, path, tokenizer, vocab):
        """
        Tokenizer and vocab are unused but provided to match superclass method signature
        """
        def vocab_to_dicts(vocab):
            t2i = dict((t, i) for i, t in enumerate(vocab))
            i2t = dict((i, t) for i, t in enumerate(vocab))
            return t2i, i2t

        # get saved processed data
        logger.debug("Loading parsed data from %s", path)
        with h5py.File(path, 'r') as f:
            self.s_vocab = f['s_vocab'][:].tolist()
            self.t_vocab = f['t_vocab'][:].tolist()
            self.s_token_to_index, self.s_index_to_token = vocab_to_dicts(self.s_vocab)
            self.t_token_to_index, self.t_index_to_token = vocab_to_dicts(self.t_vocab)
            X = f['X'][:]
            y = f['y'][:]
        self.nclass = len(self.t_vocab)

        # Trim subset and patial minibatch
        if self.subset_pct < 100:
            X = X[:int(X.shape[0] * self.subset_pct / 100.), :]
            y = y[:int(y.shape[0] * self.subset_pct / 100.), :]
            logger.debug("subset %d%% of data", self.subset_pct*100)

        extra_sentences = X.shape[0] % self.be.bsz
        if extra_sentences:
            X = X[:-extra_sentences, :]
            y = y[:-extra_sentences, :]
            logger.debug("removing %d extra sentences", extra_sentences)
        self.nbatches = X.shape[0] // self.be.bsz
        self.ndata = self.nbatches * self.be.bsz  # no leftovers

        return X, y


class Shakespeare(Dataset):
    """
    Shakespeare data set from http://cs.stanford.edu/people/karpathy/char-rnn.
    """
    def __init__(self, timesteps, path='.'):
        url = 'http://cs.stanford.edu/people/karpathy/char-rnn'
        super(Shakespeare, self).__init__('shakespeare_input.txt',
                                          url,
                                          4573338,
                                          path=path)
        self.timesteps = timesteps

    def load_data(self):
        self.filepath = self.load_zip(self.filename, self.size)
        return self.filepath

    def gen_iterators(self):
        self.load_data()
        train_path, valid_path = Text.create_valid_file(self.filepath)
        self._data_dict = {}
        self._data_dict['train'] = Text(self.timesteps, train_path)
        vocab = self._data_dict['train'].vocab
        self._data_dict['valid'] = Text(self.timesteps, valid_path, vocab=vocab)
        return self._data_dict


class PTB(Dataset):
    """
    Penn Treebank data set from http://arxiv.org/pdf/1409.2329v5.pdf

    Arguments:
        timesteps (int): number of timesteps to embed the data
        onehot_input (bool):
        tokenizer (str): name of the tokenizer function within this
                         class to use on the data
    """
    def __init__(self, timesteps, path='.',
                 onehot_input=True,
                 tokenizer=None,
                 reverse_target=False,
                 get_prev_target=False):
        url = 'https://raw.githubusercontent.com/wojzaremba/lstm/master/data'
        self.filemap = {'train': 5101618,
                        'test': 449945,
                        'valid': 399782}
        keys = list(self.filemap.keys())
        filenames = [self.gen_filename(phase) for phase in keys]
        sizes = [self.filemap[phase] for phase in keys]
        super(PTB, self).__init__(filenames,
                                  url,
                                  sizes,
                                  path=path)
        self.timesteps = timesteps
        self.onehot_input = onehot_input
        self.tokenizer = tokenizer
        if tokenizer is not None:
            assert hasattr(self, self.tokenizer)
            self.tokenizer_func = getattr(self, self.tokenizer)
        else:
            self.tokenizer_func = None

        self.reverse_target = reverse_target
        self.get_prev_target = get_prev_target

    @staticmethod
    def newline_tokenizer(s):
        """
        Tokenizer which breaks on newlines.

        Arguments:
            s (str): String to tokenize.

        Returns:
            str: String with "<eos>" in place of newlines.

        """
        # replace newlines with '<eos>' so that
        # the newlines count as words
        return s.replace('\n', '<eos>').split()

    @staticmethod
    def gen_filename(phase):
        """
        Filename generator.

        Arguments:
            phase(str): Phase

        Returns:
            string: ptb.<phase>.txt

        """
        return 'ptb.%s.txt' % phase

    def load_data(self):
        self.file_paths = {}
        for phase in self.filemap:
            fn = self.gen_filename(phase)
            size = self.filemap[phase]
            self.file_paths[phase] = self.load_zip(fn, size)
        return self.file_paths

    def gen_iterators(self):
        self.load_data()

        self._data_dict = {}
        self.vocab = None
        for phase in ['train', 'test', 'valid']:
            file_path = self.file_paths[phase]
            get_prev_target = self.get_prev_target if phase is 'train' else False
            self._data_dict[phase] = Text(self.timesteps,
                                          file_path,
                                          tokenizer=self.tokenizer_func,
                                          onehot_input=self.onehot_input,
                                          vocab=self.vocab,
                                          reverse_target=self.reverse_target,
                                          get_prev_target=get_prev_target)
            if self.vocab is None:
                self.vocab = self._data_dict['train'].vocab
        return self._data_dict


class HutterPrize(Dataset):
    """
    Hutter Prize data set from http://prize.hutter1.net/
    """
    def __init__(self, path='.'):
        super(HutterPrize, self).__init__('enwik8.zip',
                                          'http://mattmahoney.net/dc',
                                          35012219,
                                          path=path)

    def load_data(self):
        self.filepath = self.load_zip(self.filename, self.size)
        return self.filepath


class IMDB(Dataset):
    """
    IMDB data set from http://www.aclweb.org/anthology/P11-1015..
    """
    def __init__(self, vocab_size, sentence_length, path='.'):
        url = 'https://s3.amazonaws.com/text-datasets'
        super(IMDB, self).__init__('imdb.pkl',
                                   url,
                                   33213513,
                                   path=path)
        self.vocab_size = vocab_size
        self.sentence_length = sentence_length
        self.filepath = None

    def load_data(self):
        self.filepath = self.load_zip(self.filename, self.size)
        return self.filepath

    def gen_iterators(self):
        if self.filepath is None:
            self.load_data()

        data = pad_data(self.filepath, vocab_size=self.vocab_size,
                        sentence_length=self.sentence_length)
        (X_train, y_train), (X_test, y_test), nclass = data

        self._data_dict = {'nclass': nclass}
        self._data_dict['train'] = ArrayIterator(X_train, y_train, nclass=2)
        self._data_dict['test'] = ArrayIterator(X_test, y_test, nclass=2)
        return self._data_dict


class SICK(Dataset):
    """
    Semantic Similarity dataset from qcri.org (Semeval 2014).

    Arguments:
        path (str): path to SICK_data directory
    """
    def __init__(self, path='SICK_data/'):
        url = 'http://alt.qcri.org/semeval2014/task1/data/uploads/'
        self.filemap = {'train': 87341,
                        'test_annotated': 93443,
                        'trial': 16446}
        keys = list(self.filemap.keys())
        self.zip_paths = None
        self.file_paths = [self.gen_filename(phase) for phase in keys]
        self.sizes = [self.filemap[phase] for phase in keys]

        super(SICK, self).__init__(filename=self.file_paths,
                                   url=url,
                                   size=self.sizes,
                                   path=path)

    @staticmethod
    def gen_zipname(phase):
        """
        Zip filename generator.

        Arguments:
            phase(str): Phase of training/evaluation

        Returns:
            string: sick_<phase>.zip
        """
        return "sick_{}.zip".format(phase)

    @staticmethod
    def gen_filename(phase):
        """
        Filename generator for the extracted zip files.

        Arguments:
            phase(str): Phase of training/evaluation

        Returns:
            string: SICK_<phase>.txt
        """
        return "SICK_{}.txt".format(phase)

    def load_data(self):
        """
        Conditional data loader will download and extract zip files if not found locally.
        """
        self.zip_paths = {}
        for phase in self.filemap:
            zn = self.gen_zipname(phase)
            size = self.filemap[phase]
            self.zip_paths[phase] = self.load_zip(zn, size)
        return self.zip_paths

    def load_eval_data(self):
        """
        Load the SICK semantic-relatedness dataset. Data is a tab-delimited txt file,
        in the format: Sentence1\tSentence2\tScore. Data is downloaded and extracted
        from zip files if not found in directory specified by self.path.

        Returns:
            tuple of tuples of np.array: three tuples containing A & B sentences
                for train, dev, and text, along with a fourth tuple containing
                the scores for each AB pair.
        """
        if self.zip_paths is None:
            self.load_data()

        trainA, trainB, devA, devB, testA, testB = [], [], [], [], [], []
        trainS, devS, testS = [], [], []

        with open(self.path + self.gen_filename('train'), 'rb') as f:
            for line in f:
                text = line.strip().split(b'\t')
                trainA.append(text[1])
                trainB.append(text[2])
                trainS.append(text[3])

        with open(self.path + self.gen_filename('trial'), 'rb') as f:
            for line in f:
                text = line.strip().split(b'\t')
                devA.append(text[1])
                devB.append(text[2])
                devS.append(text[3])

        with open(self.path + self.gen_filename('test_annotated'), 'rb') as f:
            for line in f:
                text = line.strip().split(b'\t')
                testA.append(text[1])
                testB.append(text[2])
                testS.append(text[3])

        trainS = [float(s) for s in trainS[1:]]
        devS = [float(s) for s in devS[1:]]
        testS = [float(s) for s in testS[1:]]

        return ((np.array(trainA[1:]), np.array(trainB[1:])),
                (np.array(devA[1:]), np.array(devB[1:])),
                (np.array(testA[1:]), np.array(testB[1:])),
                (np.array(trainS), np.array(devS), np.array(testS)))



# ----------------------------------------------------------------------------
# Copyright 2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from pycuda.compiler import SourceModule
from pycuda.tools import context_dependent_memoize

"""
Binary CUDA kernels.
"""


# XNOR GEMM adapted from https://github.com/MatthieuCourbariaux/BinaryNet,
# licensed under the BSD 3-clause license. See details at
# https://github.com/MatthieuCourbariaux/BinaryNet/blob/master/LICENSE.txt
#
# A is shape (m,n), B is shape (n,k) and C is shape (m,k)
# 1 bit operations are simulated by packing 32 bits into a single integer
# and then performing bitwise operations. Matrice dimensions correspond to
# packed integers.
@context_dependent_memoize
def XNOR_gemm():

    code = r"""
#define BLOCK_SIZE 16
__global__ void xnor_gemm(unsigned int* A, unsigned int* B, float* C, int m, int n, int k) {
    // Block row and column
    int blockRow = blockIdx.y;
    int blockCol = blockIdx.x;

    // Thread row and column within Csub
    int row = threadIdx.y;
    int col = threadIdx.x;

    // Each thread block computes one sub-matrix Csub of C
    float* Csub = &C[BLOCK_SIZE * k * blockRow + BLOCK_SIZE * blockCol];

    // Shared memory used to store Asub and Bsub respectively
    __shared__ unsigned int As[BLOCK_SIZE][BLOCK_SIZE];
    __shared__ unsigned int Bs[BLOCK_SIZE][BLOCK_SIZE];

    // Each thread computes one element of Csub
    // by accumulating results into Cvalue
    // BLOCK_SIZE = 16 -> 256 threads, one per Csub element
    unsigned int Cvalue = 0;

    // Loop over all the sub-matrices of A and B that are
    // required to compute Csub
    // Multiply each pair of sub-matrices together
    // and accumulate the results
    for (int i = 0; i < (n / BLOCK_SIZE); ++i) {

        // Get sub-matrix Asub of A
        unsigned int* Asub = &A[BLOCK_SIZE * blockRow * n + BLOCK_SIZE * i];

        // Get sub-matrix Bsub of B
        unsigned int* Bsub = &B[BLOCK_SIZE * k * i + BLOCK_SIZE * blockCol];

        // Load Asub and Bsub from device memory to shared memory
        // Each thread loads one element of each sub-matrix
        As[row][col] = Asub[row*n+col];
        Bs[row][col] = Bsub[row*k+col];

        // Synchronize to make sure the sub-matrices are loaded
        // before starting the computation
        __syncthreads();

        // Multiply Asub and Bsub together
        for (int j = 0; j < BLOCK_SIZE; ++j) {
             Cvalue += __popc(As[row][j]^Bs[j][col]);
        }

        // Synchronize to make sure that the preceding
        // computation is done before loading two new
        // sub-matrices of A and B in the next iteration
        __syncthreads();
    }

    // Write Csub to device memory
    // Each thread writes one element
    if(col + blockCol * BLOCK_SIZE < k && row + blockRow * BLOCK_SIZE < m) {
        Csub[row * k + col] = -(2 * (float) Cvalue - 32 * n);
    }
}
"""

    module = SourceModule(code)
    kernel = module.get_function("xnor_gemm")
    sig = "3P 3I"
    kernel.prepare(sig)
    return kernel


def pack():

    code = r"""
__device__ unsigned int pack(float* array) {
    unsigned int rvalue = 0;
    unsigned int sign;

    for (int i = 0; i < 32; i++) {
        sign = (array[i] >= 0);
        rvalue = rvalue | (sign << i);
    }

    return rvalue;
}
"""

    return code


@context_dependent_memoize
def pack_rows():

    code = pack() + r"""
__global__ void pack_rows(float *a, unsigned int *b, int size) {
    int i = blockIdx.x * blockDim.x + threadIdx.x;

    if (i < size) {
        b[i] = pack(&a[i * 32]);
    }
}
"""

    module = SourceModule(code)
    kernel = module.get_function("pack_rows")
    sig = "2P I"
    kernel.prepare(sig)
    return kernel


@context_dependent_memoize
def pack_cols():

    code = r"""
__global__ void pack_cols(float *a, unsigned int *b, int m, int n) {
    int j = blockIdx.x * blockDim.x + threadIdx.x;
    int i = blockIdx.y * blockDim.y + threadIdx.y;

    if (j < n && 32 * i < m) {
        float num;
        unsigned int rvalue = 0;
        unsigned int sign;

        for(int k = 0; k < 32; k++) {
            num = a[j + n * (32 * i + k)];
            sign = (num >= 0);
            rvalue = rvalue | (sign << k);
        }
        b[j + n * i] = rvalue;
    }
}
"""

    module = SourceModule(code)
    kernel = module.get_function("pack_cols")
    sig = "2P 2I"
    kernel.prepare(sig)
    return kernel


def shift_element():

    code = r"""
__device__ float shift_element(float a, float b, bool value) {
    float result;

    int expb;
    if (value && b == 0) {
        return 0;
    } else if (value) {
        expb = round(log2(abs(b)));
    } else {
        expb = b;
    }

    int expa;
    double mantissa = frexp(a, &expa);
    result = ldexp(mantissa, expa + expb);

    if (value && b < 0) result = -result;

    return result;
}
"""

    return code


@context_dependent_memoize
def shift():

    code = shift_element() + r"""
__global__ void shift(
    float *a, float *b, float *c, bool value, int sizea, int b_rows, int b_cols)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;

    if (i < sizea) {
        float bi;
        if (b_rows > 1 && b_cols > 1) {
            bi = b[i];
        } else if (b_rows > 1 && b_cols == 1) {
            int step = sizea/b_rows;
            bi = b[i/step];
        } else if (b_rows == 1 && b_cols > 1) {
            bi = b[i % b_cols];
        } else if (b_rows == 1 && b_cols == 1) {
            bi = b[0];
        }

        c[i] = shift_element(a[i], bi, value);
    }
}
"""

    module = SourceModule(code)
    kernel = module.get_function("shift")
    sig = "3P 4I"
    kernel.prepare(sig)
    return kernel



# ----------------------------------------------------------------------------
# Copyright 2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
# ------------------------------------------------------------------
# Fast R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see fast-rcnn/LICENSE for details]
# Written by Ross Girshick
# ------------------------------------------------------------------
from neon.backends.util.source_module import SourceModule
from pycuda.tools import context_dependent_memoize

"""
CUDA kernels for ROI pooling layers.
There is a fprop function, a bprop function.
The fprop and bprop CUDA-C code are adapted from Fast R-CNN model.
Each of the kernels uses templating to perform %(type)
conversion so it works for all data types (currently fp32 and fp16 are supported).
"""


def map_string2func(funcname, clss):
    """
    Helper function that converts string function names to function calls
    """
    if funcname == "fprop_roipooling":
        return _get_fprop_roipooling(clss)
    if funcname == "bprop_roipooling":
        return _get_bprop_roipooling(clss)
    raise AttributeError("kernel type '" + funcname + "' not understood")


# This section of the code contains templated CUDA-C code for the kernels.
@context_dependent_memoize
def _get_fprop_roipooling(clss):

    code = r"""
#define FLT_MAX 3.402823466E+38F

__global__ void fprop_roipooling(const int nthreads,
    const int num_rois, const int img_count,
    const int channels, const int height, const int width,
    const int pooled_height, const int pooled_width,
    const float* bottom_data, const float* bottom_rois, float* top_data,
    int* argmax_data, const float spatial_scale) {
    for (int index = blockIdx.x * blockDim.x + threadIdx.x; \
        index < (nthreads); index += blockDim.x * gridDim.x){
        // (c, ph, pw, n) is an element in the pooled output
        int n = index % num_rois;
        int pw = (index / num_rois) % pooled_width;
        int ph = (index / num_rois / pooled_width) % pooled_height;
        int c = index / num_rois / pooled_width / pooled_height;

        bottom_rois += n * 5;
        int roi_batch_ind = bottom_rois[0];
        int roi_start_w = round(bottom_rois[1] * spatial_scale);
        int roi_start_h = round(bottom_rois[2] * spatial_scale);
        int roi_end_w = round(bottom_rois[3] * spatial_scale);
        int roi_end_h = round(bottom_rois[4] * spatial_scale);

        // Force malformed ROIs to be 1x1
        int roi_width = max(roi_end_w - roi_start_w + 1, 1);
        int roi_height = max(roi_end_h - roi_start_h + 1, 1);
        float bin_size_h = static_cast<float>(roi_height)
                           / static_cast<float>(pooled_height);
        float bin_size_w = static_cast<float>(roi_width)
                           / static_cast<float>(pooled_width);

        int hstart = static_cast<int>(floor(static_cast<float>(ph)
                                            * bin_size_h));
        int wstart = static_cast<int>(floor(static_cast<float>(pw)
                                            * bin_size_w));
        int hend = static_cast<int>(ceil(static_cast<float>(ph + 1)
                                         * bin_size_h));
        int wend = static_cast<int>(ceil(static_cast<float>(pw + 1)
                                         * bin_size_w));

        // Add roi offsets and clip to input boundaries
        hstart = min(max(hstart + roi_start_h, 0), height);
        hend = min(max(hend + roi_start_h, 0), height);
        wstart = min(max(wstart + roi_start_w, 0), width);
        wend = min(max(wend + roi_start_w, 0), width);
        bool is_empty = (hend <= hstart) || (wend <= wstart);

        // Define an empty pooling region to be zero
        float maxval = is_empty ? 0 : -FLT_MAX;
        // If nothing is pooled, argmax = -1 causes nothing to be backprop'd
        int maxidx = -1;

        bottom_data += c * height * width * img_count;

        for (int h = hstart; h < hend; ++h) {
          for (int w = wstart; w < wend; ++w) {
            int bottom_index = h * width * img_count + w * img_count + roi_batch_ind;
            if (bottom_data[bottom_index] > maxval) {
              maxval = bottom_data[bottom_index];
              maxidx = bottom_index;
            }
          }
        }
        top_data[index] = maxval;
        argmax_data[index] = maxidx;
        // Notice the maxidx (from bottom_index) is relative to the dimension
        // (h, w, img_count) of the feature map, so max value is HWN
    }
}

"""

    module = SourceModule(code)
    kernel = module.get_function("fprop_roipooling")
    sig = "8I 4P 1f"
    kernel.prepare(sig)
    return kernel


# This section of the code contains templated CUDA-C code for the kernels.
@context_dependent_memoize
def _get_bprop_roipooling(clss):

    code = r"""
__global__ void bprop_roipooling(const int nthreads,
    const int num_rois, const int img_count,
    const int channels, const int height, const int width,
    const int pooled_height, const int pooled_width,
    const float* top_diff, const float* bottom_rois, float* bottom_diff,
    const int* argmax_data, const float spatial_scale) {
    for (int index = blockIdx.x * blockDim.x + threadIdx.x; \
        index < (nthreads); index += blockDim.x * gridDim.x){
        // (c, h, w, n) coords in bottom data on feature map
        int n = index % img_count;
        int w = (index / img_count) % width;
        int h = (index / img_count / width) % height;
        int c = index / img_count/ width / height;

        float gradient = 0;
        // Accumulate gradient over all ROIs that pooled this element
        for (int roi_n = 0; roi_n < num_rois; ++roi_n) {
          const float* offset_bottom_rois = bottom_rois + roi_n * 5;
          int roi_batch_ind = offset_bottom_rois[0];
          // Skip if ROI's batch index doesn't match n
          if (n != roi_batch_ind) {
            continue;
          }

          int roi_start_w = round(offset_bottom_rois[1] * spatial_scale);
          int roi_start_h = round(offset_bottom_rois[2] * spatial_scale);
          int roi_end_w = round(offset_bottom_rois[3] * spatial_scale);
          int roi_end_h = round(offset_bottom_rois[4] * spatial_scale);

          // Skip if ROI doesn't include (h, w)
          const bool in_roi = (w >= roi_start_w && w <= roi_end_w &&
                               h >= roi_start_h && h <= roi_end_h);
          if (!in_roi) {
            continue;
          }

          int offset = c * pooled_height * pooled_width * num_rois;
          const float* offset_top_diff = top_diff + offset;
          const int* offset_argmax_data = argmax_data + offset;

          // Compute feasible set of pooled units that could have pooled
          // this bottom unit

          // Force malformed ROIs to be 1x1
          int roi_width = max(roi_end_w - roi_start_w + 1, 1);
          int roi_height = max(roi_end_h - roi_start_h + 1, 1);

          float bin_size_h = static_cast<float>(roi_height)
                             / static_cast<float>(pooled_height);
          float bin_size_w = static_cast<float>(roi_width)
                             / static_cast<float>(pooled_width);

          int phstart = floor(static_cast<float>(h - roi_start_h) / bin_size_h);
          int phend = ceil(static_cast<float>(h - roi_start_h + 1) / bin_size_h);
          int pwstart = floor(static_cast<float>(w - roi_start_w) / bin_size_w);
          int pwend = ceil(static_cast<float>(w - roi_start_w + 1) / bin_size_w);

          phstart = min(max(phstart, 0), pooled_height);
          phend = min(max(phend, 0), pooled_height);
          pwstart = min(max(pwstart, 0), pooled_width);
          pwend = min(max(pwend, 0), pooled_width);

          for (int ph = phstart; ph < phend; ++ph) {
            for (int pw = pwstart; pw < pwend; ++pw) {
              int top_index = ph * pooled_width * num_rois + pw * num_rois + roi_n;
              int bottom_index = h * width * img_count + w * img_count + roi_batch_ind;
              if (offset_argmax_data[top_index] == bottom_index) {
                gradient += offset_top_diff[top_index];
              }
            }
          }
        }
        bottom_diff[index] = gradient;
    }
}

"""

    module = SourceModule(code)
    kernel = module.get_function("bprop_roipooling")
    sig = "8I 4P 1f"
    kernel.prepare(sig)
    return kernel



# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Command line argument parser for neon deep learning library

This is a wrapper around the configargparse ArgumentParser class.
It adds in the default neon command line arguments and allows
additional arguments to be added using the argparse library
methods.  Lower priority defaults can also be read from a configuration file
(specified by the -c command line argument).
"""
import configargparse
import logging
from logging.handlers import RotatingFileHandler
import numpy as np
import os
import inspect

from neon import __version__ as neon_version
from neon.backends import gen_backend
from neon.backends.backend import Backend
from neon.backends.util.check_gpu import get_compute_capability, get_device_count
from neon.backends.util.check_mkl import get_mkl_lib
from neon.callbacks.callbacks import Callbacks

logger = logging.getLogger(__name__)


def extract_valid_args(args, func, startidx=0):
    """
    Given a namespace of argparser args, extract those applicable to func.

    Arguments:
        args (Namespace): a namespace of args from argparse
        func (Function): a function to inspect, to determine valid args
        startidx (int): Start index

    Returns:
        dict of (arg, value) pairs from args that are valid for func
    """
    func_args = inspect.getargspec(func).args[startidx:]
    return dict((k, v) for k, v in list(vars(args).items()) if k in func_args)


class NeonArgparser(configargparse.ArgumentParser):
    """
    Setup the command line arg parser and parse the
    arguments in sys.arg (or from configuration file).  Use the parsed
    options to configure the logging module.

    Arguments:
        desc (String) : Docstring from the calling function. This will be used
                        for the description of the command receiving the
                        arguments.
    """
    def __init__(self, *args, **kwargs):
        self._PARSED = False
        self.work_dir = os.path.join(os.path.expanduser('~'), 'nervana')
        if 'default_config_files' not in kwargs:
            kwargs['default_config_files'] = [os.path.join(self.work_dir,
                                                           'neon.cfg')]
        if 'add_config_file_help' not in kwargs:
            # turn off the auto-generated config help for config files since it
            # referenced unsettable config options like --version
            kwargs['add_config_file_help'] = False

        self.defaults = kwargs.pop('default_overrides', dict())
        super(NeonArgparser, self).__init__(**kwargs)

        # ensure that default values are display via --help
        self.formatter_class = configargparse.ArgumentDefaultsHelpFormatter

        self.setup_default_args()

    def setup_default_args(self):
        """
        Setup the default arguments used by neon
        """

        self.add_argument('--version', action='version', version=neon_version)
        self.add_argument('-c', '--config',
                          is_config_file=True,
                          help='Read values for these arguments from the '
                               'configuration file specified here first.')
        self.add_argument('-v', '--verbose', action='count',
                          default=self.defaults.get('verbose', 1),
                          help="verbosity level.  Add multiple v's to "
                               "further increase verbosity")
        # we store the negation of no_progress_bar in args.progress_bar during
        # parsing
        self.add_argument('--no_progress_bar',
                          action="store_true",
                          help="suppress running display of progress bar and "
                               "training loss")

        bm_grp = self.add_argument_group('benchmark')
        bm_grp.add_argument('--profile', action='store_true')
        bm_grp.add_argument('--profiling_method', type=str, default='time')
        bm_grp.add_argument('--profile_inference', action='store_true')
        bm_grp.add_argument('--profile_iterations', type=int, default=50)
        bm_grp.add_argument('--profile_iter_skip', type=int, default=5)

        # runtime specifc options
        rt_grp = self.add_argument_group('runtime')
        rt_grp.add_argument('-w', '--data_dir',
                            default=os.path.join(self.work_dir, 'data'),
                            help='working directory in which to cache '
                                 'downloaded and preprocessed datasets')
        rt_grp.add_argument('-e', '--epochs', type=int,
                            default=self.defaults.get('epochs', 10),
                            help='number of complete passes over the dataset to run')
        rt_grp.add_argument('-s', '--save_path', type=str,
                            default=self.defaults.get('save_path'),
                            help='file path to save model snapshots')
        rt_grp.add_argument('--serialize', nargs='?', type=int,
                            default=self.defaults.get('serialize', 0),
                            const=1, metavar='N',
                            help='serialize model every N epochs')
        rt_grp.add_argument('--model_file', help='load model from pkl file')
        rt_grp.add_argument('-l', '--log', dest='logfile', nargs='?',
                            const=os.path.join(self.work_dir, 'neon_log.txt'),
                            help='log file')
        rt_grp.add_argument('-o', '--output_file',
                            default=self.defaults.get('output_file', None),
                            help='hdf5 data file for metrics computed during '
                                 'the run, optional.  Can be used by nvis for '
                                 'visualization.')
        rt_grp.add_argument('-eval', '--eval_freq', type=int,
                            default=self.defaults.get('eval_freq', None),
                            help='frequency (in epochs) to test the eval set.')
        rt_grp.add_argument('-H', '--history', type=int,
                            default=self.defaults.get('history', 1),
                            help='number of checkpoint files to retain')
        rt_grp.add_argument('--log_token', type=str,
                            default='',
                            help='access token for data logging in real time')
        rt_grp.add_argument('--manifest', action='append', help="manifest files")
        rt_grp.add_argument('--manifest_root', type=str, default=None,
                            help='Common root path for relative path items in the '
                                 'supplied manifest files')
        be_grp = self.add_argument_group('backend')
        be_grp.add_argument('-b', '--backend', choices=Backend.backend_choices(),
                            default='gpu' if get_compute_capability() >= 3.0
                                    else 'mkl' if get_mkl_lib()
                                    else 'cpu',
                            help='backend type. Multi-GPU support is a premium '
                                 'feature available exclusively through the '
                                 'Nervana cloud. Please contact '
                                 'info@nervanasys.com for details.')
        be_grp.add_argument('-i', '--device_id', type=int,
                            default=self.defaults.get('device_id', 0),
                            help='gpu device id (only used with GPU backend)')
        be_grp.add_argument('-m', '--max_devices', type=int,
                            default=self.defaults.get('max_devices', get_device_count()),
                            help='max number of GPUs (only used with mgpu backend')

        be_grp.add_argument('-r', '--rng_seed', type=int,
                            default=self.defaults.get('rng_seed', None),
                            metavar='SEED',
                            help='random number generator seed')
        be_grp.add_argument('-u', '--rounding',
                            const=True,
                            type=int,
                            nargs='?',
                            metavar='BITS',
                            default=self.defaults.get('rounding', False),
                            help='use stochastic rounding [will round to BITS number '
                                 'of bits if specified]')
        be_grp.add_argument('-d', '--datatype', choices=['f16', 'f32', 'f64'],
                            default=self.defaults.get('datatype', 'f32'),
                            metavar='default datatype',
                            help='default floating point '
                            'precision for backend [f64 for cpu only]')
        be_grp.add_argument('-z', '--batch_size', type=int,
                            default=self.defaults.get('batch_size', 128),
                            help='batch size')
        be_grp.add_argument('--caffe', action='store_true',
                            help='match caffe when computing conv and pool layer output '
                                 'sizes and dropout implementation')
        be_grp.add_argument('--deterministic', action='store_true',
                            help='Use deterministic kernels where applicable')
        return

    def add_yaml_arg(self):
        """
        Add the yaml file argument, this is needed for scripts that
        parse the model config from yaml files

        """
        # yaml configuration file
        self.add_argument('yaml_file',
                          type=configargparse.FileType('r'),
                          help='neon model specification file')

    def add_argument(self, *args, **kwargs):
        """
        Method by which command line arguments are added to the parser.  Passed
        straight through to parent add_argument method.

        Arguments:
            *args:
            **kwargs:
        """
        if self._PARSED:
            logger.warn('Adding arguments after arguments were parsed = '
                        'may need to rerun parse_args')
            # reset so warning only comes once
            self._PARSED = False

        super(NeonArgparser, self).add_argument(*args, **kwargs)
        return

    # we never use this alias from ConfigArgParse, but defining this here
    # prevents documentation indent warnings
    def add(self):
        """ Ignored. """
        pass

    # we never use this alias from ConfigArgParse, but defining this here
    # prevents documentation indent warnings
    def add_arg(self):
        """ Ignored. """
        pass

    def parse_args(self, gen_be=True):
        """
        Parse the command line arguments and setup neon
        runtime environment accordingly

        Arguments:
            gen_be (bool): if False, the arg parser will not
                           generate the backend

        Returns:
            namespace: contains the parsed arguments as attributes

        """
        args = super(NeonArgparser, self).parse_args()
        err_msg = None  # used for relaying exception to logger

        # set up the logging
        # max thresh is 50 (critical only), min is 10 (debug or higher)
        try:
            log_thresh = max(10, 40 - args.verbose * 10)
        except (AttributeError, TypeError):
            # if defaults are not set or not -v given
            # for latter will get type error
            log_thresh = 30
        args.log_thresh = log_thresh

        # logging formater
        fmtr = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        # get the parent logger for neon
        main_logger = logging.getLogger('neon')
        main_logger.setLevel(log_thresh)

        # setup a console stderr log handler
        stderrlog = logging.StreamHandler()
        stderrlog.setFormatter(fmtr)

        # expand any user directories in paths
        for path in ['data_dir', 'save_path', 'model_file', 'output_file',
                     'logfile']:
            if getattr(args, path):
                setattr(args, path, os.path.expanduser(getattr(args, path)))

        if args.logfile:
            # add log to file as well
            filelog = RotatingFileHandler(filename=args.logfile, mode='w',
                                          maxBytes=10000000, backupCount=5)
            filelog.setFormatter(fmtr)
            filelog.setLevel(log_thresh)
            main_logger.addHandler(filelog)

            # if a log file is specified and progress bar displayed,
            # log only errors to console.
            if args.no_progress_bar:
                stderrlog.setLevel(log_thresh)
            else:
                stderrlog.setLevel(logging.ERROR)
        else:
            stderrlog.setLevel(log_thresh)

        # add this handler instead
        main_logger.propagate = False
        main_logger.addHandler(stderrlog)

        # need to write out float otherwise numpy
        # generates type in bytes not bits (f16 == 128 bits)
        args.datatype = 'float' + args.datatype[1:]
        args.datatype = np.dtype(args.datatype).type

        # invert no_progress_bar meaning and store in args.progress_bar
        args.progress_bar = not args.no_progress_bar

        if args.backend == 'cpu' and args.rounding > 0:
            err_msg = 'CPU backend does not support stochastic rounding'
            logger.exception(err_msg)
            raise NotImplementedError(err_msg)

        # done up front to avoid losing data due to incorrect path
        if args.save_path:
            savedir = os.path.dirname(os.path.abspath(args.save_path))
            if not os.access(savedir, os.R_OK | os.W_OK):
                try:
                    os.makedirs(savedir)
                except OSError:
                    err_msg = 'Can not create save_path %s' % (savedir)
            if os.path.exists(args.save_path):
                logger.warning('save file %s exists, attempting to overwrite' % args.save_path)
                if not os.access(args.save_path, os.R_OK | os.W_OK):
                    err_msg = 'Can not write to save_path file %s' % args.save_path
            if err_msg:
                logger.exception(err_msg)
                raise IOError(err_msg)

        if (args.serialize > 0) and (args.save_path is None):
            args.save_path = "neon_model.pkl"
            logger.warn('No path given for model serialization, using default "%s"',
                        args.save_path)
        if (args.save_path is not None) and (args.serialize == 0):
            args.serialize = 1
            logger.warn('No schedule given for model serialization, using default %d',
                        args.serialize)

        if args.model_file:
            err_msg = None
            if not os.path.exists(args.model_file):
                err_msg = 'Model file %s not present' % args.model_file
            if not os.access(args.model_file, os.R_OK):
                err_msg = 'No read access for model file %s' % args.model_file
            if err_msg:
                logger.exception(err_msg)
                raise IOError(err_msg)

        if args.caffe:
            args.compat_mode = 'caffe'
        else:
            args.compat_mode = None

        if args.deterministic:
            logger.warn('--deterministic flag is deprecated.  Specify random seed for '
                        'deterministic behavior.')
        # extended parsers may need to generate backend after argparsing
        if gen_be:
            # generate the backend
            gen_backend(backend=args.backend,
                        rng_seed=args.rng_seed,
                        device_id=args.device_id,
                        batch_size=args.batch_size,
                        datatype=args.datatype,
                        max_devices=args.max_devices,
                        compat_mode=args.compat_mode)

        # display what command line / config options were set (and from where)
        logger.info(self.format_values())

        if args.manifest:
            args.manifest = {k: v for k, v in [ss.split(':') for ss in args.manifest]}

        self._PARSED = True
        self.args = args
        args.callback_args = extract_valid_args(args, Callbacks.__init__, startidx=1)
        return args



# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
'''
neon visualization script.

Visualize data generated by running neon with --output_file.

For example these commands run an experiment in neon and collect metrics, then
use nvis to visualize them.

    neon <yaml path> -o mnist.h5
    nvis -i mnist.h5 -o mnist.html

HDF5 currently does not support reading while open for writing. An upcoming HDF5
release should support this officially, but in the meantime neon flushes the output
file once per epoch which generally leaves the file in a good state for periodic
inspection with nvis.

For details of all the command line arguments run this script with
the --help option.
'''
import os
import logging
import argparse
from bokeh import plotting
from neon.visualizations.figure import cost_fig, hist_fig, deconv_summary_page
from neon.visualizations.data import h5_cost_data, h5_hist_data, h5_deconv_data

logging.basicConfig(level=20)
logger = logging.getLogger()


def parse_args():
    """
    Parse command line args.
    """
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('--axis', choices=['epoch', 'minibatch'], default='epoch',
                        help='display time plots on either an epoch or minibatch axis.'
                             ' default=epoch')

    static_grp = parser.add_argument_group('Static HTML Generator')

    static_grp.add_argument('--in_file', '-i', required=True,
                            help='input hdf5 file containing data generated by running'
                                 ' an experiment in neon')

    static_grp.add_argument('--out_dir', '-o', default=os.getcwd() + "/nvis/",
                            help='output folder to write visualizations (./nvis/ by default)')

    return parser.parse_args()


def static_plot(args, plot_height=400, plot_width=600):
    """
    Generate a static html plot with train and validation cost if present in file.
    Deconvolution plots for each layer will also be generated if deconvolution data is present.
    """
    if not os.path.exists(args.out_dir):
        os.makedirs(args.out_dir)

    epoch_axis = args.axis == 'epoch'
    cost_data = h5_cost_data(args.in_file, epoch_axis)
    hist_data = h5_hist_data(args.in_file, epoch_axis)

    figs = []
    figs.append(cost_fig(cost_data, plot_height, plot_width, epoch_axis=epoch_axis))
    shared_x_range = figs[0].x_range
    if hist_data:
        for hist in hist_data:
            figs.append(hist_fig(hist, plot_height // 2, plot_width,
                                 x_range=shared_x_range, epoch_axis=epoch_axis))

    plotting.output_file(os.path.join(args.out_dir, "cost-hist.html"))
    plotting.save(plotting.vplot(*figs))

    # deconv plots
    deconv_data = h5_deconv_data(args.in_file)
    if deconv_data is not None:
        deconv_summary_page(os.path.join(args.out_dir, "deconv.html"), cost_data, deconv_data)


if __name__ == "__main__":
    args = parse_args()
    if args.in_file and args.out_dir:
        static_plot(args)



#!/usr/bin/env python
# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Script to convert old batch_writer dataset cache global means to new format.
Old format is a full image mean while the new format has a single mean value
for each channel of the input image.  This script will overwrite the existing
data cache pickle file.

Arguments:
    data cache file: path to the data cache file, this path will be printed
                     in the exception raised by neon if it detects that the
                     data cache global mean is of the old format
"""
import argparse
import os
import numpy as np
from neon.util.persist import load_obj, save_obj
from neon import logger as neon_logger

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('cache_file', help='path to data cache file')
    args = parser.parse_args()

    cache_file = args.cache_file

    # check for RW access to file
    assert os.path.exists(cache_file), 'file does not exist %s' % cache_file
    if not os.access(os.path.abspath(cache_file), os.R_OK | os.W_OK):
        raise IOError('Need to add read and/or write permissions on file %s' % cache_file)

    dc = load_obj(cache_file)

    if 'global_mean' not in dc or 'img_size' not in dc:
        raise ValueError('data cache file missing global_mean key')

    sz = dc['img_size']
    gm = dc['global_mean']

    if len(gm.shape) != 2 or (gm.shape[0] != sz * sz * 3 or gm.shape[1] != 1):
        raise ValueError('global mean shape {} does not match format expected'.format(gm.shape))

    # Collapse the full tensor mean into channel means and correct the order (RGB <-> BGR)
    dc['global_mean'] = np.mean(gm.reshape(3, -1), axis=1).reshape(3, 1)[::-1]

    save_obj(dc, cache_file)

    neon_logger.display('%s updated to new format' % cache_file)
